package core;

import java.io.File;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

//This class handles various configuration settings that can optionally be specified in the config file
public class ConfigMgr {
	private char[] 	 regionShortcuts = { 'o',   'u',             'c',           'A',       'a',       'L',       'l',       'S',       's',        'f',      'h',     'n',     'b',     't',     'd',     'G',         'g',           'K',          'k',         'e',         'x'};
	private String[] regionNames = { "Occluder","UpperClothes","LowerClothes","LeftArm","RightArm","LeftLeg", "RightLeg","LeftShoe","RightShoe","Face",   "Hair",  "Neck",    "Bag", "Hat",   "Dress", "LeftGlove", "RightGlove", "LeftSock", "RightSock", "Sunglasses","BadSegment"};
	private int[]    regionColors = { 0x0000AA,  0xFF5500,       0x0055FF,       0x00AAFF,  0x00FFFF,  0x55FFAA,  0xAAFF55,  0xFFFF00,  0xFFAA00,   0x0000FF,0xFF0000,0x008000,0x000080,0x800000,0x000055, 0x005500,   0x550000,       0x005555,   0x555500,     0xAA0000,     0xFFFFFF};

	//                                   0      1       2       3       4       5            6            7         8            9         10        11      12       13        14      15       16       17           18       19           20        21
	private String[] keypointNames   = {"Nose","L_Eye","R_Eye","L_Ear","R_Ear","L_Shoulder","L_Elbow",   "L_Wrist","R_Shoulder","R_Elbow","R_Wrist","L_Hip","L_Knee","L_Ankle","R_Hip","R_Knee","R_Ankle","L_Toes",   "R_Toes", "B_Head", "M_Shoulder","M_Hip"};
	private char[]   keypointShortcuts={'y',   'h',     'g',   'u',     't',    'w',        's',         'x',      'q',         'a',      'z',      'r',    'f',     'v',      'e',    'd',     'c',      'p',         'o',     'j',      '\0',       '\0'};
	private int[]    keypointParents = {-1,     2,      -1,    -1,      -1,      8,          5,           6,        14,          8,        9,        5,      11,      12,       11,     14,      15,       13,          16,      -1,       8,          20};
	double[][] segLength=            {{ -1,     -1,     -1,    -1,      -1,     120,         96,          80.5,       -1,        96,      80.5,      -1,     144,     130,       67,    144,     130,      45,          45,      -1,       -1,        153}, // Male
                            		  { -1,     -1,     -1,    -1,      -1,     110,         96,          80.5,       -1,        96,      80.5,      -1,     144,     130,       67,    144,     130,      45,          45,      -1,       -1,        153}, // Female
                            		  { -1,     -1,     -1,    -1,      -1,     115,         96,          80.5,       -1,        96,      80.5,      -1,     144,     130,       67,    144,     130,      45,          45,      -1,       -1,        153}}; // Child
	private boolean[]keypointIsHelper={false, false,   false,   false,  false,  false,       false,       false,    false,       false,    false,    false,  false,   false,    false,  false,   false,    false,      false,    false,    true,       true};
	
	private String annotationListName = "annotation_list.txt";

	private String  datasetName="person";
	private String[] subcategoryNames   = {"male","female","baby"};
	private String	rootDirectory=null;
	
	public String	getRootDirectory() { return rootDirectory; }
	public char[]   getRegionShortcuts() { return regionShortcuts; }
	public String[] getRegionNames() { return regionNames; }
	public int[]    getRegionColors() { return regionColors; }

	public String[] getKeypointNames() { return keypointNames; }
	public char[]   getKeypointShortcuts() { return keypointShortcuts; }
	public int[]    getKeypointParents() { return keypointParents; }
	public boolean[]getKeypointIsHelper() { return keypointIsHelper; }
	
	public double getSegmentLength(int subcategory, int keypointID) { return segLength[subcategory][keypointID]; }
	
	public String   datasetName() { return datasetName; }
	public String[] getSubcategoryNames() { return subcategoryNames; }
	public String getAnnotationListName() { return annotationListName; }
	
	public int getKeypointID(String kpName) {
		if (kpName!=null)
			for (int i=0; i<getKeypointNames().length; ++i)
				if (kpName.equals(getKeypointNames()[i]))
					return i;
		return -1;
	}
	
	public ConfigMgr() {}	// default instantiation

	public ConfigMgr(String configFile) {
		File f = new File(configFile);
		Document doc;
		try {
			DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
			doc = docBuilder.parse (f);
			NodeList rsList = doc.getElementsByTagName("Dataset");
			assert(rsList.getLength()<2);
			if (rsList.getLength()==1)
				datasetName = ((Element)rsList.item(0)).getAttribute("name");

			rsList = doc.getElementsByTagName("RootDirectory");
			assert(rsList.getLength()<2);
			if (rsList.getLength()==1) {
				rootDirectory = ((Element)rsList.item(0)).getAttribute("value");
				if (!rootDirectory.endsWith("/"))
					rootDirectory+="/";
			}

			rsList = doc.getElementsByTagName("AnnotationListName");
			assert(rsList.getLength()<2);
			if (rsList.getLength()==1) {
				annotationListName = ((Element)rsList.item(0)).getAttribute("value");
			}

			rsList = doc.getElementsByTagName("Subcategory");
			subcategoryNames = new String[rsList.getLength()];
			segLength = new double[Math.max(1, subcategoryNames.length)][];
			for (int i=0; i<rsList.getLength(); ++i) {
				Element rs = (Element)rsList.item(i);
				subcategoryNames[i] = rs.getAttribute("name");
			}
			
			rsList = doc.getElementsByTagName("Region");
			regionShortcuts = new char[rsList.getLength()];
			regionNames = new String[rsList.getLength()];
			regionColors = new int[rsList.getLength()];
			for (int i=0; i<rsList.getLength(); ++i) {
				Element rs = (Element)rsList.item(i);
				regionNames[i] = rs.getAttribute("name");
				String val = rs.getAttribute("key");
				if (val.length()!=1)
					throw new IllegalArgumentException("String instead of char");
				regionShortcuts[i] = val.charAt(0);
				val = rs.getAttribute("color");
				regionColors[i] = Integer.parseInt(val,16);
			}
			
			rsList = doc.getElementsByTagName("Keypoint");
			keypointNames = new String[rsList.getLength()];
			keypointShortcuts = new char[rsList.getLength()];
			keypointParents = new int[rsList.getLength()];
			keypointIsHelper = new boolean[rsList.getLength()];
			String[] parentNames = new String[rsList.getLength()];

			for (int i=0; i<segLength.length; ++i)
				segLength[i] = new double[keypointNames.length];

			for (int i=0; i<rsList.getLength(); ++i) {
				Element rs = (Element)rsList.item(i);
				keypointNames[i] = rs.getAttribute("name");
				if (rs.hasAttribute("key")) {
					String val = rs.getAttribute("key");
					if (val.length()!=1)
						throw new IllegalArgumentException("String instead of char");
					keypointShortcuts[i] = val.charAt(0);
				} else
					keypointShortcuts[i] = 0;
					
				if (rs.hasAttribute("parent"))
					parentNames[i] = rs.getAttribute("parent");
				
				double length=-1;
				if (rs.hasAttribute("length"))
					length = Double.parseDouble(rs.getAttribute("length"));
						
				for (int j=0; j<segLength.length; ++j) {
					if (subcategoryNames.length>0 && rs.hasAttribute(subcategoryNames[j]))
						segLength[j][i] = Double.parseDouble(rs.getAttribute(subcategoryNames[j]));
					else
						segLength[j][i] = length;
				}
				keypointIsHelper[i] = false;
				if (rs.hasAttribute("helperKeypoint") && Boolean.valueOf(rs.getAttribute("helperKeypoint"))) {
					keypointIsHelper[i]=true;
				}
			}
			for (int i=0; i<rsList.getLength(); ++i)
				if (parentNames[i]==null)
					keypointParents[i] = -1;
				else {
					keypointParents[i] = getKeypointID(parentNames[i]);
					assert(keypointParents[i]>=0);
				}
		} catch (SAXParseException err) {
			System.out.println ("** Parsing error" + ", line " 
					+ err.getLineNumber () + ", uri " + err.getSystemId ());
			System.out.println(" " + err.getMessage ());
			return;
		} catch (SAXException e) {
			Exception x = e.getException ();
			((x == null) ? e : x).printStackTrace ();
			return;
		} catch (Throwable t) {
			t.printStackTrace ();
			return;
		}
	}
}
