package core;

import java.awt.AWTEvent;
import java.awt.event.KeyEvent;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import javax.media.j3d.Appearance;
import javax.media.j3d.BoundingBox;
import javax.media.j3d.BoundingSphere;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Canvas3D;
import javax.media.j3d.ColoringAttributes;
import javax.media.j3d.GeometryArray;
import javax.media.j3d.LineArray;
import javax.media.j3d.LineAttributes;
import javax.media.j3d.RenderingAttributes;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.TransparencyAttributes;
import javax.media.j3d.WakeupCriterion;
import javax.media.j3d.WakeupOnAWTEvent;
import javax.media.j3d.WakeupOr;
import javax.vecmath.AxisAngle4d;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

import com.sun.j3d.utils.behaviors.vp.OrbitBehavior;
import com.sun.j3d.utils.geometry.Cylinder;
import com.sun.j3d.utils.geometry.Sphere;
import com.sun.j3d.utils.pickfast.behaviors.PickMouseBehavior;
import com.sun.j3d.utils.universe.ViewingPlatform;


public class Human3DPose extends PickMouseBehavior {
		
	static class T {
		// Maps the keypoint names in config file to some hard-coded keypoint names for people
		// This can be written more elegantly using an enum but this way makes the rest of the code simpler
		public static int Nose,L_Eye,R_Eye,L_Ear,R_Ear,L_Shoulder,L_Elbow,L_Wrist,R_Shoulder,R_Elbow,R_Wrist,L_Hip,L_Knee,L_Ankle,R_Hip,R_Knee,R_Ankle,L_Toes,R_Toes,B_Head,M_Shoulder,M_Hip;

		public static boolean init(ConfigMgr configMgr) {
			Nose = configMgr.getKeypointID("Nose");             if (Nose<0) return false;
			L_Eye = configMgr.getKeypointID("L_Eye");           if (L_Eye<0) return false;
			R_Eye = configMgr.getKeypointID("R_Eye");           if (R_Eye<0) return false;
			L_Ear = configMgr.getKeypointID("L_Ear");           if (L_Ear<0) return false;
			R_Ear = configMgr.getKeypointID("R_Ear");           if (R_Ear<0) return false;
			L_Toes = configMgr.getKeypointID("L_Toes"); if (L_Toes<0) return false;
			R_Toes = configMgr.getKeypointID("R_Toes"); if (R_Toes<0) return false;
			L_Shoulder = Math.max(configMgr.getKeypointID("L_Shoulder"), configMgr.getKeypointID("F_L_Hip")); if (L_Shoulder<0) return false;
			R_Shoulder = Math.max(configMgr.getKeypointID("R_Shoulder"), configMgr.getKeypointID("F_R_Hip")); if (R_Shoulder<0) return false;
			L_Elbow = Math.max(configMgr.getKeypointID("L_Elbow"), configMgr.getKeypointID("F_L_Knee")); if (L_Elbow<0) return false;
			R_Elbow = Math.max(configMgr.getKeypointID("R_Elbow"), configMgr.getKeypointID("F_R_Knee")); if (R_Elbow<0) return false;
			L_Wrist = Math.max(configMgr.getKeypointID("L_Wrist"), configMgr.getKeypointID("F_L_Hoof")); if (L_Wrist<0) return false;
			R_Wrist = Math.max(configMgr.getKeypointID("R_Wrist"), configMgr.getKeypointID("F_R_Hoof")); if (R_Wrist<0) return false;
			L_Hip = Math.max(configMgr.getKeypointID("L_Hip"), configMgr.getKeypointID("B_L_Hip")); if (L_Hip<0) return false;
			R_Hip = Math.max(configMgr.getKeypointID("R_Hip"), configMgr.getKeypointID("B_R_Hip")); if (R_Hip<0) return false;
			L_Knee = Math.max(configMgr.getKeypointID("L_Knee"), configMgr.getKeypointID("B_L_Knee")); if (L_Knee<0) return false;
			R_Knee = Math.max(configMgr.getKeypointID("R_Knee"), configMgr.getKeypointID("B_R_Knee")); if (R_Knee<0) return false;
			L_Ankle = Math.max(configMgr.getKeypointID("L_Ankle"), configMgr.getKeypointID("B_L_Hoof")); if (L_Ankle<0) return false;
			R_Ankle = Math.max(configMgr.getKeypointID("R_Ankle"), configMgr.getKeypointID("B_R_Hoof")); if (R_Ankle<0) return false;
			B_Head = configMgr.getKeypointID("B_Head");         if (B_Head<0) return false;
			M_Shoulder = Math.max(configMgr.getKeypointID("M_Shoulder"), configMgr.getKeypointID("F_M_Hip")); if (M_Shoulder<0) return false;
			M_Hip = Math.max(configMgr.getKeypointID("M_Hip"), configMgr.getKeypointID("B_M_Hip")); if (M_Hip<0) return false;
			return true;
		}
	}
	
	
	private KeypointMgr keypoints;
	private TransformGroup root;
	private WakeupCriterion[] wakeupEvents;
	private WakeupOr allWakeupEvents;
	private boolean suppressKeyEvents = false;
	private boolean humanPoseCanBeEnabled;
	
	enum SegmentType    { ShoulderLine, Spine,   HipLine, UpperLeftArm, LowerLeftArm, UpperRightArm, LowerRightArm, UpperLeftLeg, LowerLeftLeg, UpperRightLeg, LowerRightLeg, RightFoot, LeftFoot };
	// segK is the beginning of the segment, segK.parent is the end
	private int[] segK;//  = { T.L_Shoulder, T.M_Hip, T.R_Hip, T.L_Elbow,    T.L_Wrist,     T.R_Elbow,    T.R_Wrist,     T.L_Knee,     T.L_Ankle,    T.R_Knee,      T.R_Ankle, T.R_Toes, T.L_Toes};
	
	private OrbitBehavior orbit;
	private UIManager uiMgr;
	private ConfigMgr configMgr;
	private ViewingPlatform viewPlatform;
	public Transform3D normalizedViewXform = new Transform3D();
	private PoseOptimizer poseOptimizer;

	private Vector<Segment> segments;
	private Vector<Joint> joints;
	private Head head;
	
	class Head  {
		private TransformGroup world2Neck, neck2HeadCtr, headCtr2Nose, headCtr2LEye, headCtr2REye;
		private Sphere headS, lEyeS, rEyeS, noseS;
		private Joint lShoulder, rShoulder, neck, lEye, rEye, lEar, rEar, nose;
				
		public Head(TransformGroup root, Joint lShoulder, Joint rShoulder, Joint neck, Joint lEye, Joint rEye, Joint lEar, Joint rEar, Joint nose) {
			this.lShoulder = lShoulder;
			this.rShoulder = rShoulder;
			this.neck = neck;
			this.lEye = lEye;
			this.rEye = rEye;
			this.lEar = lEar;
			this.rEar = rEar;
			this.nose = nose;
			
			world2Neck = new TransformGroup();
			world2Neck.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
			root.addChild(world2Neck);
			
			neck2HeadCtr = new TransformGroup();
			neck2HeadCtr.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
			world2Neck.addChild(neck2HeadCtr);
			
			headCtr2Nose = new TransformGroup();
			neck2HeadCtr.addChild(headCtr2Nose);

			headCtr2LEye = new TransformGroup();
			neck2HeadCtr.addChild(headCtr2LEye);

			headCtr2REye = new TransformGroup();
			neck2HeadCtr.addChild(headCtr2REye);

			Transform3D t = new Transform3D();
			Transform3D rxn = new Transform3D();
			Transform3D rzn = new Transform3D();
			Transform3D t3n = new Transform3D();
			Transform3D sc = new Transform3D();
			
			double noseAngleZ = 0;
			double noseAngleX = Math.PI*0.5;

			rzn.setRotation(new AxisAngle4d(0,0,1,noseAngleZ));
			rxn.setRotation(new AxisAngle4d(1,0,0,noseAngleX));
			t3n.setTranslation(new Vector3d(0,1,0));
			sc.setScale(new Vector3d(1,1,1.5));

			t.set(rzn);
			t.mul(rxn);
			t.mul(t3n);
			t.mul(sc);
			headCtr2Nose.setTransform(t);

			double eyeAngleZ = Math.PI*0.25;
			double eyeAngleX = Math.PI*0.35;

			rzn.setRotation(new AxisAngle4d(0,0,1,eyeAngleZ));
			rxn.setRotation(new AxisAngle4d(1,0,0,eyeAngleX));
			t3n.setTranslation(new Vector3d(0,1,0));

			t.set(rzn);
			t.mul(rxn);
			t.mul(t3n);
			headCtr2REye.setTransform(t);

			rzn.setRotation(new AxisAngle4d(0,0,1,-eyeAngleZ));
			t.set(rzn);
			t.mul(rxn);
			t.mul(t3n);
			headCtr2LEye.setTransform(t);

			
			Appearance ap = new Appearance();
			ap.setCapability(Appearance.ALLOW_RENDERING_ATTRIBUTES_READ);
			ColoringAttributes ca = new ColoringAttributes();
			ca.setColor(0,0,1);
			ap.setColoringAttributes(ca);
			RenderingAttributes ra = new RenderingAttributes();
			ra.setVisible(false);
			ra.setCapability(RenderingAttributes.ALLOW_VISIBLE_WRITE);
			ap.setRenderingAttributes(ra);

			Appearance eyeAp = new Appearance();
			eyeAp.setCapability(Appearance.ALLOW_RENDERING_ATTRIBUTES_READ);
			ca = new ColoringAttributes();
			ca.setColor(1,1,1);
			eyeAp.setColoringAttributes(ca);
			
			ra = new RenderingAttributes();
			ra.setVisible(false);
			ra.setCapability(RenderingAttributes.ALLOW_VISIBLE_WRITE);
			eyeAp.setRenderingAttributes(ra);

			lEyeS = new Sphere(0.1f, Sphere.ENABLE_APPEARANCE_MODIFY, eyeAp);
			headCtr2LEye.addChild(lEyeS);
			rEyeS = new Sphere(0.1f, Sphere.ENABLE_APPEARANCE_MODIFY, eyeAp);
			headCtr2REye.addChild(rEyeS);
			noseS = new Sphere(0.1f, Sphere.ENABLE_APPEARANCE_MODIFY, eyeAp);
			headCtr2Nose.addChild(noseS);
			headS = new Sphere(1, Sphere.ENABLE_APPEARANCE_MODIFY, ap);
			neck2HeadCtr.addChild(headS);
		}
						
		public void update(double minScale) {
			boolean headVisible = neck.isActive() && lShoulder.isActive() && rShoulder.isActive();
			boolean faceVisible = headVisible  && lEye.isActive() && rEye.isActive() && nose.isActive();
			
			if (headVisible) {
				Vector3d shoulderVec = new Vector3d(lShoulder.x()-rShoulder.x(), lShoulder.y()-rShoulder.y(), lShoulder.z()-rShoulder.z());
				double sLen = shoulderVec.length();
				shoulderVec.normalize();
				
				Transform3D t3 = new Transform3D();
				Transform3D sc = new Transform3D();
				Transform3D rot = new Transform3D();
				
				t3.setTranslation(new Vector3d(neck.x(),neck.y(),neck.z()));
				sc.setScale(sLen*0.2);
				rot.setRotation(new AxisAngle4d(0, shoulderVec.z, shoulderVec.y, -Math.acos(shoulderVec.x)));

				Transform3D t = new Transform3D();
				t.mul(t3);
				t.mul(sc);
//				t.mul(rot);
				world2Neck.setTransform(t);
				
				double N = 1;
				double headAngleY = 0;//Math.PI/4.0*0;
				double headAngleZ = 0;//Math.PI/4.0*0;
				double headAngleX = 0;//Math.PI/4.0;

				if (lEye.isActive() && rEye.isActive()) {
					double dx = lEye.x()-rEye.x();
					double dy = lEye.y()-rEye.y();
					if (dx!=0 || dy!=0) {
						headAngleZ = Math.atan2(dy, dx);
						if (nose.isActive()) {
							// Estimate head Y rotate
	
							double lambda = ((lEye.x()-nose.x())*dx + (lEye.y()-nose.y())*dy) / (dx*dx + dy*dy);
/*	
							double eyeDist = Math.sqrt(dx*dx + dy*dy);
							dx = (lEye.x()+rEye.x())/2 - nose.x();
							dy = (lEye.y()+rEye.y())/2 - nose.y();
							double noseMideyeDist = Math.sqrt(dx*dx + dy*dy);
							double rel = eyeDist / noseMideyeDist;		
							*/
	//						System.out.println("Lambda: "+lambda+ " rel="+rel + " eyeDist="+eyeDist/minScale);
	//						if (eyeDist/minScale > 11) {
								// Mostly frontal face.
								headAngleY = -Math.PI*(Math.min(Math.max(lambda,0),1)-0.5);
	//							System.out.println("hax="+headAngleX);
	//						}
						}
					} else
						faceVisible=false;
				}
								
				
				Transform3D rx = new Transform3D();
				Transform3D rz = new Transform3D();
				Transform3D ry = new Transform3D();
				rz.setRotation(new AxisAngle4d(0,0,1,headAngleZ));
				rx.setRotation(new AxisAngle4d(1,0,0,headAngleX));
				ry.setRotation(new AxisAngle4d(0,1,0,headAngleY));
				t3.setTranslation(new Vector3d(0,N,0));
												
				t.setIdentity();
				t.mul(rz);
				t.mul(rx);
				t.mul(ry);
				t.mul(t3);
				neck2HeadCtr.setTransform(t);
				
			}
			
			lEyeS.getAppearance().getRenderingAttributes().setVisible(faceVisible);
			rEyeS.getAppearance().getRenderingAttributes().setVisible(faceVisible);
			noseS.getAppearance().getRenderingAttributes().setVisible(faceVisible);
			headS.getAppearance().getRenderingAttributes().setVisible(headVisible);
		}
	}

	
	class Joint  {
		private TransformGroup xform;
		private Sphere sphere;
		public Keypoint k;
		private double sphereScale=1;
		
		public double x() { return k.x(); }
		public double y() { return k.y(); }
		public double z() { return k.z(); }

		public boolean isActive() { return k.isActive(); }
		public Keypoint.ZOrder getZOrder() { return k.getZOrder(); }
		public String toString() { return k.toString()+" z="+z(); };
				
		public Joint(TransformGroup root, Keypoint k, float radius) {
			this.k = k;
			Transform3D tr = new Transform3D();
			tr.setTranslation(new Vector3d(x(),y(),z()));
			xform = new TransformGroup(tr);
			xform.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
			
			Appearance ap = new Appearance();
			ap.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_WRITE);
			ap.setCapability(Appearance.ALLOW_RENDERING_ATTRIBUTES_WRITE);
			ap.setCapability(Appearance.ALLOW_RENDERING_ATTRIBUTES_READ);
			ColoringAttributes ca = new ColoringAttributes();
			ca.setColor(0,0,1);
			ap.setColoringAttributes(ca);
			RenderingAttributes ra = new RenderingAttributes();
			ra.setVisible(false);
			ra.setCapability(RenderingAttributes.ALLOW_VISIBLE_WRITE);
			ap.setRenderingAttributes(ra);
			sphere = new Sphere(radius,Sphere.ENABLE_APPEARANCE_MODIFY,ap);
			xform.addChild(sphere);
			root.addChild(xform);
		}
		
		public void setScale(double s) { 
			sphereScale = s;
			update();
		}
		
		public void setZ(double z) {
			k.setZ(z);
			update();
		}
		
		public void update() {
			Transform3D tr = new Transform3D();
			tr.setTranslation(new Vector3d(x(),y(),z()));
			Transform3D sc = new Transform3D();
			sc.setScale(sphereScale);
			tr.mul(sc);
			xform.setTransform(tr);
			sphere.getAppearance().getRenderingAttributes().setVisible(isActive());
		}
	}

	class Segment {
		public Joint k,k_parent;
		private Shape3D shape;
		private TransformGroup cylinderXform;
		private Cylinder cylinder;
		private double length;
		private Point3d tmp = new Point3d();
		private SegmentType type;

		public boolean isActive() { return k.isActive() && k_parent.isActive(); }
		public void setVisible(boolean visible) {
			shape.getAppearance().getRenderingAttributes().setVisible(visible);
		}
		public void setLength(double length) { this.length = length; } // from setBodyType
		public SegmentType type() { return type; }
		
		public void setColor(float r, float g, float b) {
			ColoringAttributes ca = new ColoringAttributes();
			ca.setColor(r,g,b);
			shape.getAppearance().setColoringAttributes(ca);
		}
		
		public void setLinePattern(int pattern) {
			shape.getAppearance().getLineAttributes().setLinePattern(pattern);
		}
		
		public Segment(SegmentType type, TransformGroup root, Joint k, Joint kParent, double length) {
			this.type = type;
			this.length = length;
			this.k = k;
			this.k_parent = kParent;
			LineArray xGeom = new LineArray(2, GeometryArray.COORDINATES);
			xGeom.setCapability(LineArray.ALLOW_COORDINATE_WRITE);

			shape =  new Shape3D(xGeom);
			Appearance ap = new Appearance();
			ap.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_WRITE);
			ap.setCapability(Appearance.ALLOW_RENDERING_ATTRIBUTES_WRITE);
			ap.setCapability(Appearance.ALLOW_RENDERING_ATTRIBUTES_READ);
			ap.setCapability(Appearance.ALLOW_LINE_ATTRIBUTES_READ);
			ColoringAttributes ca = new ColoringAttributes();
			ca.setColor(0,0,1);
			ap.setColoringAttributes(ca);
			RenderingAttributes ra = new RenderingAttributes();
			ra.setVisible(false);
			ra.setCapability(RenderingAttributes.ALLOW_VISIBLE_WRITE);
			ap.setRenderingAttributes(ra);
			LineAttributes la = new LineAttributes(5, LineAttributes.PATTERN_SOLID,true);
			la.setCapability(LineAttributes.ALLOW_PATTERN_WRITE);
			
			ap.setLineAttributes(la);
			shape.setAppearance(ap);
			shape.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
			shape.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
			shape.setCapability(Shape3D.ALLOW_GEOMETRY_READ);			
			root.addChild(shape);
			
			cylinderXform = new TransformGroup();
			cylinderXform.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
			root.addChild(cylinderXform);
			cylinder = new Cylinder(0.1f,1);
			TransparencyAttributes ta = new TransparencyAttributes(TransparencyAttributes.BLENDED,0.5f,
				    TransparencyAttributes.BLEND_SRC_ALPHA,
				    TransparencyAttributes.BLEND_ONE);
			ap.setTransparencyAttributes(ta);
			cylinder.setAppearance(ap);
			cylinderXform.addChild(cylinder);
		}

		public double getScale() {
			if (!isActive()) return 0;
			double length2DSqrd = (k.x()-k_parent.x())*(k.x()-k_parent.x()) + (k.y()-k_parent.y())*(k.y()-k_parent.y());	
			return Math.sqrt(length2DSqrd)/length;
		}
		
		// Updates its coordinates from the keypoints. Invisible if the keypoints are not available
		public void setCoordinates(double minScale) {
			if (isActive()) {
				LineArray xGeom = (LineArray)shape.getGeometry();
				tmp.set(k.x(), k.y(), k.z()); xGeom.setCoordinate(0, tmp);
				tmp.set(k_parent.x(), k_parent.y(), k_parent.z()); xGeom.setCoordinate(1, tmp);

				double length3d = getDirection(tmp);
								
				Vector3d ctr = new Vector3d((k.x()+k_parent.x())/2, (k.y()+k_parent.y())/2, (k.z()+k_parent.z())/2);
				
				Transform3D t3 = new Transform3D();
				Transform3D sc = new Transform3D();
				Transform3D rot = new Transform3D();
				
				t3.setTranslation(ctr);
				double scale = minScale*100;
				if (type==SegmentType.Spine)
					scale*=2;
				sc.setScale(new Vector3d(scale,length3d,scale));
				rot.setRotation(new AxisAngle4d(tmp.z, 0, -tmp.x, Math.acos(tmp.y)));
				

				Transform3D t = new Transform3D();
				t.mul(t3);
				t.mul(rot);
				t.mul(sc);

				try {
				cylinderXform.setTransform(t);
				} catch (Exception e) {
					System.out.println("Exception thrown by setTransform: "+e);
				}
			} 
			setVisible(isActive());
		}
		
		// Returns normalized direction from parent to child keypoint
		public double getDirection(Point3d dir) {
			tmp.set(k.x()-k_parent.x(), k.y()-k_parent.y(), k.z()-k_parent.z());
			double length3d = Math.sqrt(tmp.x*tmp.x + tmp.y*tmp.y + tmp.z*tmp.z);
			tmp.scale(1/length3d);	// normalize
			dir.set(tmp);
			return length3d;
		}


		// Assuming the first point has correct coordinates, sets the coordinates of the second one
		public double propagateZ(double minScale, double parentZ, double perspZ) {
			if (!isActive())
				return Double.NaN;
			
			assert(k_parent.z() == parentZ);
						
			double scaledLength = length*minScale;//*   perspZ/(perspZ+z);

			double length2DSqrd = (k.x()-k_parent.x())*(k.x()-k_parent.x()) + (k.y()-k_parent.y())*(k.y()-k_parent.y());	
			double rel_z = Math.sqrt(Math.max(0, scaledLength*scaledLength - length2DSqrd));

			if (isFrontoParallel()) {
				float v = (float)Math.pow(rel_z/scaledLength,0.5); 
				setColor(v,0,1-v);	
				if (k.k.getLabel()==T.R_Hip)
					rel_z = -rel_z;	// When the hips and shoulders are shortened, make sure they turn the same way
			} else {
				setColor(0,0,1);
			}
			
			if (k.getZOrder()==Keypoint.ZOrder.Far)
				rel_z = - rel_z;

			double z = parentZ + rel_z;

			k.setZ(z);
			return z;
		}
		
		// How much does the segment need to shrink to result in no error
		public boolean isFrontoParallel() { return k.getZOrder()==Keypoint.ZOrder.Equal; }
	}

	public Human3DPose(KeypointMgr keypoints, Canvas3D canvas, ViewingPlatform viewPlatform, BranchGroup branch, UIManager uiMgr, ConfigMgr configMgr) {
		super(canvas, branch, new BoundingSphere());
		this.configMgr = configMgr;
		humanPoseCanBeEnabled = T.init(configMgr);
	    segK  = new int[]{ T.L_Shoulder, T.M_Hip, T.R_Hip, T.L_Elbow,    T.L_Wrist,     T.R_Elbow,    T.R_Wrist,     T.L_Knee,     T.L_Ankle,    T.R_Knee,      T.R_Ankle,  T.R_Toes, T.L_Toes};

		this.keypoints=keypoints;
		this.viewPlatform = viewPlatform;
		this.uiMgr = uiMgr;
		viewPlatform.getViewPlatformTransform().getTransform(normalizedViewXform);
		
		root = new TransformGroup();
		root.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		branch.addChild(root);

		orbit = new OrbitBehavior(canvas, OrbitBehavior.DISABLE_TRANSLATE | OrbitBehavior.DISABLE_ZOOM | OrbitBehavior.REVERSE_ALL);
		orbit.setSchedulingBounds(new BoundingSphere(new Point3d(0.0, 0.0, 0.0), 100.0));
		viewPlatform.setViewPlatformBehavior(orbit);

		if (!humanPoseCanBeEnabled)
			return;
		
		int[] noSpheres = {T.M_Hip, T.Nose, T.L_Eye, T.R_Eye, T.L_Ear, T.R_Ear, T.B_Head };
		Arrays.sort(noSpheres);
		
		joints = new Vector<Joint>();
		for (int i=0; i<keypoints.numKeypoints(); ++i) {
			float radius = (Arrays.binarySearch(noSpheres, i)>=0) ? 0 : 0.03f;
			joints.add(new Joint(root,keypoints.getKeypoint(i),radius));
		}

		int bodyType = Math.max(0, keypoints.getBodyType());
		segments = new Vector<Segment>();
		for (int i=0; i<SegmentType.values().length; ++i) {
			Keypoint k = keypoints.getKeypoint(segK[i]);
			Keypoint k_parent = keypoints.getKeypoint(k.getParent());
			double segment_length = configMgr.getSegmentLength(bodyType, segK[i]);
			segments.add(new Segment(SegmentType.values()[i],root,joints.get(k.getLabel()), joints.get(k_parent.getLabel()), segment_length));
		}
	
		head = new Head(root, joints.get(T.L_Shoulder), joints.get(T.R_Shoulder), 
				joints.get(T.M_Shoulder),joints.get(T.L_Eye), joints.get(T.R_Eye),
				joints.get(T.L_Ear), joints.get(T.R_Ear), joints.get(T.Nose));
		
		poseOptimizer= new PoseOptimizer(joints,segments,keypoints.numKeypoints());
	}

	public void initialize(){
		wakeupEvents = new WakeupCriterion[1];
		wakeupEvents[0] = new WakeupOnAWTEvent(KeyEvent.KEY_PRESSED);
		allWakeupEvents = new WakeupOr(wakeupEvents);
		wakeupOn(allWakeupEvents);
	}	
	
	@SuppressWarnings("unchecked")
	public void processStimulus(Enumeration criteria){
		while (criteria.hasMoreElements())
		{
			WakeupCriterion wakeup=(WakeupCriterion) criteria.nextElement();
			if (wakeup instanceof WakeupOnAWTEvent) {
				WakeupOnAWTEvent event = (WakeupOnAWTEvent)wakeup;
				AWTEvent[] events = event.getAWTEvent();

				for (int i = 0; i < events.length; ++i) {
					if (events[i] instanceof KeyEvent) {
						KeyEvent keyEvent = (KeyEvent)events[i];
						if (keyEvent.getID() == KeyEvent.KEY_PRESSED)
							processKeyPressed(keyEvent);
						continue;
					}
				}
			}
		}

		suppressKeyEvents = false;
		wakeupOn(allWakeupEvents);
	}

	private void processKeyPressed(KeyEvent ev) {
		if (suppressKeyEvents && keypoints.getKeypointFromShortcut(ev.getKeyChar())!=null)
			return;
		uiMgr.processKeyPressed(ev);
	}

	public void resetView() {
		viewPlatform.getViewPlatformTransform().setTransform(normalizedViewXform);				
		display();
		// All of this below is a workaround for a Java bug - sometimes it doesn't refresh the segments. So we have to hide them all, let it draw, and in a delayed event show them.
		if (!humanPoseCanBeEnabled)
			return;
		setSegmentVisibility(false);
	    Timer timer = new Timer();
	    timer.schedule(new ShowSegmentsTask(this), 50);
	}
	
	public void setSegmentVisibility(boolean visible) {
		for (Segment s: segments)
			if (visible)
				s.setVisible(s.isActive());
			else
				s.setVisible(false);			
	}
	
	public boolean humanPoseCanBeEnabled() {
		return humanPoseCanBeEnabled;
	}
	
	public void enable(boolean enable) {
		if (!humanPoseCanBeEnabled)
			return;
		orbit.setEnable(enable);
		suppressKeyEvents = true;
	}
	public void setBodyType(int bodyType) {
		if (!humanPoseCanBeEnabled)
			return;
		for (int s=0; s<segments.size(); ++s) {
			segments.get(s).setLength(configMgr.getSegmentLength(Math.max(0, bodyType), segments.get(s).k.k.getLabel()));//     segLength[Math.max(0, bodyType)][s]);
		}
	}
	
	private double propagateZ() {
		if (!humanPoseCanBeEnabled)
			return 0;
		// Update mid shoulder and mid hip
		Keypoint sl = keypoints.getKeypoint(T.L_Shoulder);
		Keypoint sr = keypoints.getKeypoint(T.R_Shoulder);
		Keypoint sm = keypoints.getKeypoint(T.M_Shoulder);
		if (sl.isActive() && sr.isActive()) {
			sm.changeStatus(Keypoint.StatusEnum.Active);
			sm.implicitKeypointSet((sl.x()+sr.x())/2, (sl.y()+sr.y())/2, sl.getZOrder());
		} else { // make it outside image
			sm.changeStatus(Keypoint.StatusEnum.Inactive);
		}

		Keypoint hl = keypoints.getKeypoint(T.L_Hip);
		Keypoint hr = keypoints.getKeypoint(T.R_Hip);
		Keypoint hm = keypoints.getKeypoint(T.M_Hip);
		if (hl.isActive() && hr.isActive()) {
			hm.changeStatus(Keypoint.StatusEnum.Active);
			Keypoint.ZOrder zorder=hl.getZOrder();
			if (sr.isActive()) {
				if (hl.getZOrder() == sr.getZOrder()) {
					if (hl.getZOrder()!=Keypoint.ZOrder.Equal)
						System.out.println("Error: Inconsistent torso Z order!");
					zorder=Keypoint.ZOrder.Equal;
				} else if (hl.getZOrder()==Keypoint.ZOrder.Equal)
					zorder = (sr.getZOrder()==Keypoint.ZOrder.Far) ? Keypoint.ZOrder.Close : Keypoint.ZOrder.Far;
			}
			hm.implicitKeypointSet((hl.x()+hr.x())/2, (hl.y()+hr.y())/2, zorder);
		} else { // make it outside image
			hm.changeStatus(Keypoint.StatusEnum.Inactive);
		}

		for (Segment s: segments) {
			s.setVisible(s.isActive());
		}

		double perspZ = 1000000;

		Segment maxSeg=null;
		double minScale = 0, minScaleFrontoParallel=0;
		double scaleSum=0;
		int scaleNum=0;
		for (Segment s: segments) {
			if (s.isActive()) {
				double scale = s.getScale();
				if (scale>minScale) {
					minScale = scale;
					maxSeg = s;									
				}
				if (s.isFrontoParallel()) {
					scaleSum=scaleSum+scale;
					scaleNum++;
	
					if (scale>minScaleFrontoParallel)
						minScaleFrontoParallel = scale;
				}
			}
		}
		if (scaleNum>0) {
			double avgFPMinScale = scaleSum/scaleNum;
//			if (minScale>avgFPMinScale) {
//				System.out.println("MinScale:"+minScale+" avs fpminscale="+avgFPMinScale);
//			}
			minScale = Math.max(avgFPMinScale, minScale);	// we found at least one frontoparallel segment. The min scale is their avg scale, but not smaller than a scale from non-frontoparallel
		} else {
			minScale = minScale*0.9; // we couldn't find a frontoparallel segment. Use a bent one, but make the scale smaller
			maxSeg = null;
		}

		propagateZInternal(minScale,perspZ);
		/*
		perspZ = 10000;
		double step = perspZ/2;
		while (step>1) {
			if (getEnergy(minScale, perspZ+step) < getEnergy(minScale, perspZ-step))
				perspZ = perspZ+step;
			else
				perspZ = perspZ-step;
			step=step/2;
		}
		
		for (int s=0; s<segments.size(); ++s) {
			Segment seg = segments.get(s);
			if (seg.isActive() && seg.isFrontoParallel()) {
				double preferredScale = seg.getScale()/minScale;
				double z = (seg.k.z()+seg.k_parent.z())/2;
				double newScale = preferredScale*(perspZ/(z+perspZ));
				System.out.println(SegmentType.values()[s]+" z="+z+" scale="+preferredScale+" new scale="+newScale);
			}
		}
		*/

		for (Segment s: segments) 
			s.setLinePattern(s==maxSeg ? LineAttributes.PATTERN_DASH : LineAttributes.PATTERN_SOLID);
		poseOptimizer.setMaxSeg(maxSeg);
				
		return minScale;
	}
	/*
	double getEnergy(double minScale, double perspZ) {
		double energy=0;
		for (int s=0; s<segments.size(); ++s) {
			Segment seg = segments.get(s);
			if (seg.isActive() && seg.isFrontoParallel()) {
				double preferredScale = seg.getScale(perspZ)/minScale;
				double z = (seg.k.z()+seg.k_parent.z())/2;
				double newScale = preferredScale*(perspZ/(z+perspZ));
				energy+=(1-newScale)*(1-newScale);
			}
		}
		System.out.println("perspZ = "+perspZ+" energy="+energy);
		return energy;
	}
	*/
	void propagateZInternal(double minScale, double perspZ) {
		// Set the X and Y coordinates of visible segments and compute the minimum scale
		
		Keypoint rShoulder = keypoints.getKeypointIfActive(T.R_Shoulder);
		if (rShoulder==null) {
			if (keypoints.getKeypointIfActive(T.L_Shoulder)!=null) {
				double sl_z = 0;
				joints.get(T.L_Shoulder).setZ(sl_z);
				double el_z = segments.get(SegmentType.UpperLeftArm.ordinal()).propagateZ(minScale, sl_z, perspZ);
				if (!Double.isNaN(el_z)) 
					segments.get(SegmentType.LowerLeftArm.ordinal()).propagateZ(minScale,el_z, perspZ);
				for (Segment s: segments)
					s.setCoordinates(minScale);
			}
			return;			
		}

		double sr_z = 0;

		double sl_z = segments.get(SegmentType.ShoulderLine.ordinal()).propagateZ(minScale, sr_z, perspZ);
		if (!Double.isNaN(sl_z)) {
			joints.get(T.R_Shoulder).setZ(sr_z);
			joints.get(T.M_Shoulder).setZ((sr_z+sl_z)/2);
			double el_z = segments.get(SegmentType.UpperLeftArm.ordinal()).propagateZ(minScale, sl_z, perspZ);
			if (!Double.isNaN(el_z)) 
				segments.get(SegmentType.LowerLeftArm.ordinal()).propagateZ(minScale,el_z, perspZ);

			double hm_z = segments.get(SegmentType.Spine.ordinal()).propagateZ(minScale, (sl_z+sr_z)/2, perspZ);
			if (!Double.isNaN(hm_z)) {
				// lower body active. Both hips must be active.
				// The midhip has correct Z but not the left/right hip.
				double h_diff = segments.get(SegmentType.HipLine.ordinal()).propagateZ(minScale,0, perspZ);
				assert(!Double.isNaN(h_diff));
				double hl_z = hm_z - h_diff/2;
				double hr_z = hm_z + h_diff/2;
				segments.get(SegmentType.HipLine.ordinal()).k.setZ(hr_z);
				joints.get(T.L_Hip).setZ(hl_z);
				joints.get(T.M_Hip).setZ((hr_z+hl_z)/2);


				double kl_z = segments.get(SegmentType.UpperLeftLeg.ordinal()).propagateZ(minScale, hl_z, perspZ);
				if (!Double.isNaN(kl_z)) {
					double al_z=segments.get(SegmentType.LowerLeftLeg.ordinal()).propagateZ(minScale,kl_z, perspZ);
					if (!Double.isNaN(al_z))
						segments.get(SegmentType.LeftFoot.ordinal()).propagateZ(minScale,al_z, perspZ);						
				}
				double kr_z = segments.get(SegmentType.UpperRightLeg.ordinal()).propagateZ(minScale,hr_z, perspZ);
				if (!Double.isNaN(kr_z)) {
					double ar_z=segments.get(SegmentType.LowerRightLeg.ordinal()).propagateZ(minScale,kr_z, perspZ);
					if (!Double.isNaN(ar_z))
						segments.get(SegmentType.RightFoot.ordinal()).propagateZ(minScale,ar_z, perspZ);
				}
			}
		}

		double er_z = segments.get(SegmentType.UpperRightArm.ordinal()).propagateZ(minScale, sr_z, perspZ);
		if (!Double.isNaN(er_z)) 
			segments.get(SegmentType.LowerRightArm.ordinal()).propagateZ(minScale, er_z, perspZ);

		for (Segment s: segments)
			s.setCoordinates(minScale);
	}
	
	public void display() {
		if (!humanPoseCanBeEnabled)
			return;
		
		double minScale=propagateZ();
		
		// Center and scale all the points
		
		Point3d p = new Point3d();
		BoundingBox box = null;
		for (Joint j : joints) {
			if (j.isActive()) {
				p.set(j.x(), j.y(), j.z());
				if (box==null)
					box = new BoundingBox(p,p);
				else
					box.combine(p);
			}
		}
		if (box!=null) {
			Point3d lower = new Point3d();
			box.getLower(lower);
			Point3d upper = new Point3d();
			box.getUpper(upper);
			
			Vector3d lv = new Vector3d(-(upper.x+lower.x)/2,-(upper.y+lower.y)/2,-(upper.z+lower.z)/2);
			Transform3D t3d = new Transform3D();
			t3d.setTranslation(lv);
	
			Transform3D sc = new Transform3D();
			double dist = lower.distance(upper);
	
			if (dist>0) {
				sc.setScale(2/dist);
				for (Joint j : joints)
					j.setScale(dist/2);
			}
			
		
			sc.mul(t3d);
			root.setTransform(sc);
		}
		
		for (Joint j : joints)
			j.update();
		head.update(minScale);
	}
	
	public void endIncrementalCorrections() {
		if (!humanPoseCanBeEnabled)
			return;
		poseOptimizer.endIncrementalCorrections();
	}
	
	public void fixTorso() {
		if (!humanPoseCanBeEnabled)
			return;
		poseOptimizer.fixTorso();
	}

	public void doIncrementalCorrections(boolean justMakeMaxSegBigger, boolean faster) {
		if (!humanPoseCanBeEnabled)
			return;
		poseOptimizer.doIncrementalCorrections(justMakeMaxSegBigger, faster);
	}
	
	@Override
	public void updateScene(int x, int y) {
		System.out.println("Update scene called...");
	}

}

class ShowSegmentsTask extends TimerTask {
	Human3DPose pose;
	public ShowSegmentsTask(Human3DPose pose) {
		this.pose=pose;
	}
    public void run() {
    	pose.setSegmentVisibility(true);
    }
  }

