package core;

import java.awt.image.BufferedImage;

import javax.media.j3d.Appearance;
import javax.media.j3d.GeometryArray;
import javax.media.j3d.QuadArray;
import javax.media.j3d.RenderingAttributes;
import javax.media.j3d.Shape3D;
import javax.swing.JOptionPane;
import javax.vecmath.Point2d;
import javax.vecmath.Point3d;
import javax.vecmath.TexCoord2f;

import com.sun.j3d.utils.image.TextureLoader;

public class ImageMgr {
	private Shape3D imageNode;
	private String imgName=new String();
	private int imgWidth, imgHeight;
	TextureLoader loader=null;

	public double getAspectRatio() { return imgWidth / (double)imgHeight; }
	public Shape3D getImage() { return imageNode; }
	public BufferedImage getBufferedImage() { return loader.getImage().getImage(); }
		
	public boolean setImage(String filename, boolean readOnly) {
		if (imgName.equals(filename))
			return false;
		imgName=filename;
		System.out.println("FILE: "+filename);
		loader=null;
		try {
			if (readOnly)
				loader = new TextureLoader(Thread.currentThread().getContextClassLoader().getResource(filename),"RGB",null );
			else
				loader = new TextureLoader(filename,"RGB",null);		
		} catch (java.lang.IllegalArgumentException e) {
			JOptionPane.showMessageDialog(null,"Cannot open image "+filename,"I/O Error",JOptionPane.ERROR_MESSAGE);						
		} catch (Exception e) {
			JOptionPane.showMessageDialog(null,"Cannot open image "+filename,"I/O Error",JOptionPane.ERROR_MESSAGE);			
		}
				
		imgWidth = loader.getImage().getWidth();
		imgHeight = loader.getImage().getHeight();
		double aspectRatio = getAspectRatio();

		QuadArray plane = (QuadArray)imageNode.getGeometry();

		Point3d p = new Point3d();
		
		p.set(-aspectRatio, -1,  0);	plane.setCoordinate(0, p);
		p.set(aspectRatio,  -1,  0);	plane.setCoordinate(1, p);
		p.set(aspectRatio,   1,  0);	plane.setCoordinate(2, p);
		p.set(-aspectRatio,  1,  0);	plane.setCoordinate(3, p);
		
		imageNode.getAppearance().setTexture(loader.getTexture());
		return true;
	}

	boolean isInsideImage(double ux, double uy) {
		double ar = getAspectRatio();
		return (uy>=0 && uy<=1 && ux>=-(ar-1)/2 && ux<=(ar+1)/2);
	}
	
	public Point2d pixelToUnitCoords(double x, double y) {
		double ux = x/(double)imgHeight - (getAspectRatio()-1.0)/2.0;
		double uy = 1 - y/(double)imgHeight;
		return new Point2d(ux,uy);
	}

	public Point3d unitToPixelCoords(double ux, double uy, double uz) {
		double x = (ux + (getAspectRatio()-1.0)/2.0)*imgHeight;
		double y = (1 - uy)*imgHeight;
		double z = uz*imgHeight;
		return new Point3d(x,y,z);
	}

	public ImageMgr() {
		QuadArray plane = new QuadArray(4, GeometryArray.COORDINATES | GeometryArray.TEXTURE_COORDINATE_2);

		TexCoord2f q = new TexCoord2f();
		q.set(0.0f, 0.0f);	plane.setTextureCoordinate(0, 0, q);
		q.set(1.0f, 0.0f);  plane.setTextureCoordinate(0, 1, q);
		q.set(1.0f, 1.0f);	plane.setTextureCoordinate(0, 2, q);
		q.set(0.0f, 1.0f);	plane.setTextureCoordinate(0, 3, q);
		
		RenderingAttributes ra = new RenderingAttributes();
		Appearance ap = new Appearance();
		ap.setRenderingAttributes(ra);
		ap.setCapability(Appearance.ALLOW_TEXTURE_WRITE);
		imageNode = new Shape3D(plane, ap);
		imageNode.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
		imageNode.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
		imageNode.getGeometry().setCapability(QuadArray.ALLOW_COORDINATE_WRITE);
	}
}
