package core;

import java.awt.Font;

import javax.media.j3d.Appearance;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.ColoringAttributes;
import javax.media.j3d.GeometryArray;
import javax.media.j3d.Group;
import javax.media.j3d.LineArray;
import javax.media.j3d.Node;
import javax.media.j3d.RenderingAttributes;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Switch;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Color3f;
import javax.vecmath.Point3d;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3d;

import com.sun.j3d.utils.geometry.Text2D;

public class Keypoint {
	private int label;
	private int parent;
	private double x,y,z;
	private BranchGroup keypointGroup;
	private Shape3D cross;
	private char shortcut;
	private KeypointMgr keypointMgr;
	
	public enum StatusEnum { Inactive, Active, Hilighted };
	private StatusEnum status;
	
	enum ZOrder { Close, Equal, Far };
	private ZOrder zorder = ZOrder.Equal;	// z-order of keypoint relative to its parent 		
	
	private boolean hidden=false;
	
	private TransformGroup xform;
	
	private Switch textGroup;
	private Appearance geomAppearance;

	public char shortcut() { return shortcut; }
	
	public BranchGroup getRoot() { return keypointGroup; }
	public int getLabel() { return label; }
	public String getName() { return keypointMgr.keypointName(label); }
	public int getParent() { return parent; }
	public ZOrder getZOrder() { return zorder; }
	public double x() { return x; }
	public double y() { return y; }
	public double z() { return z; }
	public Point3d getCoordinates() { return new Point3d(x,y,z); }
	public void setZ(double z) { this.z = z; } 
	public boolean isActive() { return status!=StatusEnum.Inactive; }
	
	public String toString() { return getName()+" "+zorder+" hidden="+hidden+" x="+x+" y="+y+" z="+z; }

	
	public void setZOrder(ZOrder zorder) { 
		if (this.zorder!=zorder) {
			this.zorder = zorder; 
			updateTransform(); 
			keypointMgr.setDirty(true);
		}
	}
	
	public void toggleZOrder() {
		switch (zorder) {
			case Close: setZOrder(ZOrder.Equal); return;
			case Equal: setZOrder(ZOrder.Far); return;
			case Far  : setZOrder(ZOrder.Close);
		}
	}
	
	public void setHidden(boolean hidden) { 
		if (this.hidden!=hidden) {
			this.hidden = hidden; 
			keypointMgr.setDirty(true);
			updateAppearance();
		}
	}
	
	public boolean getHidden() { return hidden; }
		
	public void changeStatus(StatusEnum newStatus) {
		if (status == newStatus)
			return;
		if (status==StatusEnum.Inactive || newStatus==StatusEnum.Inactive)
			keypointMgr.setDirty(true);
		
		status = newStatus;
		cross.getAppearance().getRenderingAttributes().setVisible(isActive());
		if (!isActive()) {
			zorder = ZOrder.Equal;
			x=0;//Double.NaN;
			y=0;//Double.NaN;
			textGroup.setWhichChild(Switch.CHILD_NONE);			
			hidden = false;
		}
			
		updateAppearance();
	}	
	
	private void updateAppearance() {
		ColoringAttributes ca = new ColoringAttributes();
		if (status==StatusEnum.Inactive) {
			textGroup.setWhichChild(Switch.CHILD_NONE);
		} else if (status==StatusEnum.Active) {
			ca.setColor(0,0,1);
			textGroup.setWhichChild(Switch.CHILD_NONE);
		} else {
			// Hilighted
			ca.setColor(1,0,0);
			textGroup.setWhichChild(hidden ? 3 : 2);
		}
		if (hidden)
			ca.setColor(0,1,0);
		geomAppearance.setColoringAttributes(ca);
	}
	
	public void moveTo(double x, double y) {		
		this.x = x;
		this.y = y;
		updateTransform();
		if (!isActive())
			changeStatus(StatusEnum.Active);

		keypointMgr.setDirty(true);
	}
	public void move(double dx, double dy) { moveTo(x+dx, y+dy); }
	
	// Used to set properties of implicit keypoints, such as the mid shoulder. Does not raise the dirty flag
	public void implicitKeypointSet(double x, double y, ZOrder zorder) {
		this.x = x;
		this.y = y;
		this.zorder = zorder;
	}
	
	void updateTransform() {		
		double scale = getLabelScale();
		
		double[] trans = {scale,0,0,x,  0,scale,0,y,  0,0,scale,KeypointMgr.keypoints_z, 0,0,0,1};
		xform.setTransform(new Transform3D(trans));
	}
	
	private Shape3D createKeypointGeometry() {
		LineArray xGeom = new LineArray(12, GeometryArray.COORDINATES);
		float s=0.04f;

		xGeom.setCoordinate(0, new Point3f(-s,0,0));
		xGeom.setCoordinate(1, new Point3f(0,s,0));

		xGeom.setCoordinate(2, new Point3f(0,s,0));
		xGeom.setCoordinate(3, new Point3f(s,0,0));

		xGeom.setCoordinate(4, new Point3f(s,0,0));
		xGeom.setCoordinate(5, new Point3f(0,-s,0));

		xGeom.setCoordinate(6, new Point3f(0,-s,0));
		xGeom.setCoordinate(7, new Point3f(-s,0,0));

		xGeom.setCoordinate(8, new Point3f(-s,0,0));
		xGeom.setCoordinate(9, new Point3f(s,0,0));

		xGeom.setCoordinate(10, new Point3f(0,-s,0));
		xGeom.setCoordinate(11, new Point3f(0,s,0));

		cross =  new Shape3D(xGeom);
		geomAppearance = new Appearance();
		geomAppearance.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_WRITE);
		geomAppearance.setCapability(Appearance.ALLOW_RENDERING_ATTRIBUTES_READ);
		ColoringAttributes ca = new ColoringAttributes();
		ca.setColor(0,0,1);
		geomAppearance.setColoringAttributes(ca);
		RenderingAttributes ra = new RenderingAttributes();
		ra.setCapability(RenderingAttributes.ALLOW_VISIBLE_WRITE);
		geomAppearance.setRenderingAttributes(ra);
		cross.setAppearance(geomAppearance);
		cross.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
		cross.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
		cross.setUserData(this);
		return cross;
	}
	
	private Node createLabelNode() {
    	textGroup = new Switch();
		textGroup.setCapability(Switch.ALLOW_SWITCH_WRITE);
		Text2D t = new Text2D(getName(), new Color3f(0.5f,0.5f,0.5f), "Times", 20, Font.PLAIN);  // gray = disabled
		t.setUserData(this);
		textGroup.addChild(t);
		t = new Text2D(getName(), new Color3f(0,0,1), "Times", 20, Font.PLAIN); // blue = active (NOT USED)
		t.setUserData(this);
		textGroup.addChild(t);
		t = new Text2D(getName(), new Color3f(1,0,0), "Times", 20, Font.PLAIN); // red = hilighted
		t.setUserData(this);
		textGroup.addChild(t);
		t = new Text2D(getName(), new Color3f(0,1,0), "Times", 20, Font.PLAIN);	// green = hidden
		t.setUserData(this);
		textGroup.addChild(t);
		textGroup.setWhichChild(0);
		return textGroup;
	}
	
	public double getLabelScale() { 
		double scale=keypointMgr.getLabelScale();
		if (zorder==ZOrder.Close)
			scale=scale*1.4;
		else if (zorder==ZOrder.Far)
			scale=scale*0.8;
		return scale;
	}
	

	Keypoint(int label, int parent, char shortcut, KeypointMgr mgr) {
		this.label = label;
		this.parent = parent;
		this.shortcut = shortcut;
		this.keypointMgr = mgr;

		keypointGroup = new BranchGroup();
		keypointGroup.setCapability(TransformGroup.ENABLE_PICK_REPORTING);		
		keypointGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		keypointGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		keypointGroup.setPickable(true);

		Transform3D t3d = new Transform3D();
		t3d.setTranslation(new Vector3d(x,y,KeypointMgr.keypoints_z));

		xform = new TransformGroup(t3d);
		xform.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		xform.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		xform.setCapability(TransformGroup.ALLOW_PICKABLE_WRITE);
		xform.setCapability(TransformGroup.ENABLE_PICK_REPORTING);
		xform.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);
		xform.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
		xform.setPickable(true);
		xform.setUserData(label);
		
		xform.addChild(createKeypointGeometry());

		t3d.setTranslation(new Vector3d(0.05,-0.03,0));
		TransformGroup xform2 = new TransformGroup(t3d);
		xform2.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
		xform2.setCapability(Group.ALLOW_CHILDREN_EXTEND);
		xform2.addChild(createLabelNode());

		xform.addChild(xform2);
		keypointGroup.addChild(xform);
		changeStatus(StatusEnum.Inactive);
	}

}
