package core;

import java.util.Vector;

import javax.media.j3d.BoundingBox;
import javax.media.j3d.Switch;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Point2d;
import javax.vecmath.Point3d;

import core.IOMgr.KeypointState;

public class KeypointMgr {	
	private Vector<Keypoint> keypoints = new Vector<Keypoint>();
	private TransformGroup root;
	private Switch rootSwitch;
	private double labelScale=0.5;
	private ImageMgr imgMgr;
	private UIManager uiMgr;
	private ConfigMgr configMgr;
	private int bodyType;
	
	
	public static final double keypoints_z = 0.1;
	
	public TransformGroup getRoot() { return root; }
	
	public void setLabelScale(double s) {
		labelScale = s;
		for (Keypoint k: keypoints)
			k.updateTransform();
	}
	
	public double getLabelScale() { return labelScale; }
	public boolean isInsideImage(double ux, double uy) { return imgMgr.isInsideImage(ux,uy); }
	public void setDirty(boolean isDirty) { if (uiMgr!=null) uiMgr.setDirty(isDirty); }
	public void setUIMgr(UIManager uiMgr) { this.uiMgr = uiMgr; }
	public void setBodyType(int bt) { bodyType = bt; }
	public int getBodyType() { return bodyType; }
	public String keypointName(int kp) { return configMgr.getKeypointNames()[kp]; }
	public int keypointID(String kpName) {
		for (int i=0; i<configMgr.getKeypointNames().length; ++i)
			if (kpName.equals(configMgr.getKeypointNames()[i]))
				return i;
		return -1;
	}

	public int numKeypoints() { return configMgr.getKeypointNames().length; }
	
	public void setVisible(boolean visible) {
		rootSwitch.setWhichChild(visible ? 0 : Switch.CHILD_NONE);
	}
		
	public BoundingBox getBounds() {
		Point3d p = new Point3d();
		BoundingBox box = null;
		for (Keypoint k : keypoints) {
			if (k.isActive()) {
				p.set(k.x(), k.y(), 0);
				if (box==null)
					box = new BoundingBox(p,p);
				else
					box.combine(p);
			}
		}
		return box;
		
	}

	public Keypoint getKeypointIfActive(int label) {
		if (label==-1)
			return null;
		Keypoint k=keypoints.get(label);
		return k.isActive() ?  k : null;
	}
	public Keypoint getKeypoint(int label) {
		if (label==-1)
			return null;
		return keypoints.get(label);
	}
	
	public Keypoint getKeypointFromShortcut(char key) {
		for (Keypoint k: keypoints)
			if (k.shortcut()==key)
				return k;
		return null;
	}
	
	public void keyPressed(char shortcut, double mouseX, double mouseY) {
		Keypoint k = getKeypointFromShortcut(shortcut);
		if (k!=null) {
			k.moveTo(mouseX,mouseY);
			k.changeStatus(Keypoint.StatusEnum.Hilighted);
		}
	}
	
	public void read(IOMgr ioMgr) {
		for (int label=0; label<configMgr.getKeypointNames().length; ++label) {
			Keypoint kp = getKeypoint(label);
			KeypointState st = ioMgr.getKeypoint(label);
			if (st==null) {
				kp.changeStatus(Keypoint.StatusEnum.Inactive);
			} else {
				kp.changeStatus(Keypoint.StatusEnum.Active);
				Point2d unitCoords = imgMgr.pixelToUnitCoords(st.x, st.y);
				kp.moveTo(unitCoords.x, unitCoords.y);
				kp.setZOrder(st.zorder);
				kp.setHidden(st.hidden);
			}
		}
		bodyType = ioMgr.getBodyType();
	}
	
	public void write(IOMgr ioMgr) {		
		ioMgr.clearKeypoints();
		for (Keypoint kp : keypoints) {
			if (kp.isActive()) {
				Point3d pixelCoords = imgMgr.unitToPixelCoords(kp.x(), kp.y(), kp.z());
				ioMgr.setKeypoint(kp.getLabel(), new KeypointState(pixelCoords.x, pixelCoords.y, pixelCoords.z, kp.getZOrder(), kp.getHidden()));
			}
		}
		ioMgr.setBodyType(bodyType);
	}

	public KeypointMgr(Switch rootSwitch, TransformGroup root, ImageMgr imgMgr, ConfigMgr configMgr) {
		this.root = root;
		this.rootSwitch = rootSwitch;
		this.imgMgr = imgMgr;
		this.configMgr = configMgr;

		for (int i=0; i<numKeypoints(); ++i) {
			Keypoint k=new Keypoint(i,configMgr.getKeypointParents()[i],configMgr.getKeypointShortcuts()[i],this);
			keypoints.add(k);
			if (configMgr.getKeypointShortcuts()[i]>0)
				root.addChild(k.getRoot());		// If it has a shortcut it is also visible
		}
	}
}
