package core;

import java.applet.Applet;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GraphicsConfiguration;
import java.awt.GridLayout;
import java.awt.Panel;
import java.awt.Toolkit;
import java.io.File;

import javax.media.j3d.BoundingSphere;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Canvas3D;
import javax.media.j3d.Switch;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.View;
import javax.swing.JFileChooser;
import javax.swing.filechooser.FileFilter;
import javax.vecmath.Vector3d;

import com.sun.j3d.utils.applet.MainFrame;
import com.sun.j3d.utils.universe.SimpleUniverse;
import com.sun.j3d.utils.universe.ViewingPlatform;

@SuppressWarnings("serial")
public class MainApp extends Applet {
	KeypointMgr keypointMgr;
	ConfigMgr configMgr;
	IOMgr ioMgr;
	UIManager uiMgr;
	ImageMgr imageMgr = new ImageMgr();
	SegmentationMgr segMgr;
	Switch taskSwitch = new Switch();	// toggle between 3D pose and segmentation
	TransformGroup pbImageXform = new TransformGroup();
	TransformGroup imageXform = new TransformGroup();
	
	BranchGroup createScene(Canvas3D canvas, ViewingPlatform viewPlatform) { 
		BranchGroup objRoot = new BranchGroup();

		TransformGroup xform1 = new TransformGroup();
		xform1.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		objRoot.addChild(xform1);
		
		TransformGroup xform2 = new TransformGroup();
		xform2.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		xform1.addChild(xform2);

		imageXform.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);		
		imageXform.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		imageXform.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);

		imageXform.addChild(imageMgr.getImage());


		Transform3D s3 = new Transform3D();		
		s3.setScale(new Vector3d(2,2,1));

		Transform3D t3 = new Transform3D();		
		t3.setTranslation(new Vector3d(-1,-1,0));
		t3.mul(s3);

		TransformGroup keypointXForm = new TransformGroup(t3);

		Switch keypointsSwitch = new Switch();	
		keypointsSwitch.addChild(keypointXForm);
		keypointsSwitch.setWhichChild(0);
		keypointsSwitch.setCapability(Switch.ALLOW_SWITCH_WRITE);
		keypointMgr = new KeypointMgr(keypointsSwitch, keypointXForm, imageMgr, configMgr);

		imageXform.addChild(keypointsSwitch);

		xform2.addChild(imageXform);

		uiMgr = new UIManager(canvas,objRoot,new BoundingSphere(), xform1,xform2,imageXform,pbImageXform,keypointMgr,viewPlatform,imageMgr,ioMgr,configMgr,taskSwitch);
		uiMgr.setSchedulingBounds(new BoundingSphere());
		keypointMgr.setUIMgr(uiMgr);
		objRoot.addChild(uiMgr);

		return objRoot;
	}
	
	BranchGroup createScene2(Canvas3D canvas, ViewingPlatform viewPlatform) { 
		BranchGroup objRoot = new BranchGroup();

		objRoot.addChild(taskSwitch);
		BranchGroup pose3DRoot = new BranchGroup();
		taskSwitch.addChild(pose3DRoot);
		
		Human3DPose poseMgr = new Human3DPose(keypointMgr,canvas, viewPlatform,pose3DRoot,uiMgr,configMgr);
		poseMgr.setSchedulingBounds(new BoundingSphere());
		pose3DRoot.addChild(poseMgr);
		uiMgr.setPoseMgr(poseMgr);

		BranchGroup segmentRoot = new BranchGroup();
		taskSwitch.addChild(segmentRoot);
		
		TransformGroup xform1 = new TransformGroup();
		xform1.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		segmentRoot.addChild(xform1);
		
		TransformGroup xform2 = new TransformGroup();
		xform2.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		xform1.addChild(xform2);

		
		xform2.addChild(pbImageXform);
		segMgr = new SegmentationMgr(canvas,objRoot,new BoundingSphere(),xform1,xform2,imageXform,pbImageXform,uiMgr,configMgr);
		segMgr.setSchedulingBounds(new BoundingSphere());
		pbImageXform.addChild(segMgr);
		pbImageXform.addChild(segMgr.getImage());
		pbImageXform.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		uiMgr.setSegMgr(segMgr);
		
		taskSwitch.setWhichChild(0);
		taskSwitch.setCapability(Switch.ALLOW_SWITCH_WRITE);

		return objRoot;
	}

	public MainApp() {		
		String databasePath = "demo_data/person/";
		boolean readOnly = true;
		final JFileChooser fc = new JFileChooser();
		if (new File("C:/hat/data/person").exists()) {
			fc.setCurrentDirectory(new File("C:/hat/data/person"));
		} else if (new File("/hat/data").exists()) {
			fc.setCurrentDirectory(new File("/hat/data/person"));
		}
	    fc.setDialogTitle("Please choose dataset config.xml or cancel for demo");
	    fc.setFileFilter(new ExtensionFileFilter("xml",new String[] { "xml", "XML" }));
	    int returnVal = fc.showOpenDialog(null);
		if( returnVal == JFileChooser.APPROVE_OPTION) {
			String configFile = fc.getSelectedFile().getAbsolutePath();
			configMgr = new ConfigMgr(configFile);
			if (configMgr.getRootDirectory()==null) {
				databasePath = fc.getCurrentDirectory().getAbsolutePath()+"/";
			} else
				databasePath = configMgr.getRootDirectory();
			readOnly=false;
		} else {
			configMgr = new ConfigMgr();
			//databasePath = fc.getCurrentDirectory().getAbsolutePath()+"/";
			//readOnly=false;
		}
		
		ioMgr=new IOMgr(databasePath,"images/","info/","segmentations/",configMgr.getAnnotationListName(),readOnly, configMgr);
		setLayout(new BorderLayout());
		Panel canvases = new Panel();
		canvases.setLayout(new GridLayout(1,2));
		
		GraphicsConfiguration config = SimpleUniverse.getPreferredConfiguration();
		{
			Canvas3D canvas = new Canvas3D(config);
			canvases.add(canvas);

			SimpleUniverse u = new SimpleUniverse(canvas);
			View v = u.getViewer().getView();
			v.setProjectionPolicy(View.PARALLEL_PROJECTION);
			u.getViewingPlatform().setNominalViewingTransform();
			u.addBranchGraph(createScene(canvas,u.getViewingPlatform()));				
		}

		{
			Canvas3D canvas = new Canvas3D(config);
			canvases.add(canvas);

			SimpleUniverse u = new SimpleUniverse(canvas);
			View v = u.getViewer().getView();
			v.setProjectionPolicy(View.PARALLEL_PROJECTION);
			v.setTransparencySortingPolicy(View.TRANSPARENCY_SORT_GEOMETRY);
			u.addBranchGraph(createScene2(canvas,u.getViewingPlatform()));		

		}
		add("Center",canvases);		
		Panel bp = new Panel();
		add("South", bp);
		uiMgr.addButtonPanel(bp);
		uiMgr.syncFromIOMgr();
	}

	public static void main(String[] args) {
		MainFrame m = new MainFrame(new MainApp(), 512, 512);	
		Dimension size = Toolkit.getDefaultToolkit ().getScreenSize();
		m.setBounds(0,0,size.width,size.height-50);
	}
}

class ExtensionFileFilter extends FileFilter {
	  String description;

	  String extensions[];

	  public ExtensionFileFilter(String description, String extension) {
	    this(description, new String[] { extension });
	  }

	  public ExtensionFileFilter(String description, String extensions[]) {
	    if (description == null) {
	      this.description = extensions[0];
	    } else {
	      this.description = description;
	    }
	    this.extensions = (String[]) extensions.clone();
	    toLower(this.extensions);
	  }

	  private void toLower(String array[]) {
	    for (int i = 0, n = array.length; i < n; i++) {
	      array[i] = array[i].toLowerCase();
	    }
	  }

	  public String getDescription() {
	    return description;
	  }

	  public boolean accept(File file) {
	    if (file.isDirectory()) {
	      return true;
	    } else {
	      String path = file.getAbsolutePath().toLowerCase();
	      for (int i = 0, n = extensions.length; i < n; i++) {
	        String extension = extensions[i];
	        if ((path.endsWith(extension) && (path.charAt(path.length() - extension.length() - 1)) == '.')) {
	          return true;
	        }
	      }
	    }
	    return false;
	  }
	}

