package core;
import java.awt.AWTEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.util.Enumeration;

import javax.media.j3d.Alpha;
import javax.media.j3d.BoundingBox;
import javax.media.j3d.BoundingSphere;
import javax.media.j3d.Bounds;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Canvas3D;
import javax.media.j3d.PositionInterpolator;
import javax.media.j3d.ScaleInterpolator;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.WakeupCriterion;
import javax.media.j3d.WakeupOnAWTEvent;
import javax.media.j3d.WakeupOr;
import javax.vecmath.Point2d;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

import com.sun.j3d.utils.picking.behaviors.PickMouseBehavior;

// Implements zoom and pan. Used for the left image and right segmentation image.

public abstract class PanZoomMgr extends PickMouseBehavior {
	protected Canvas3D canvas;		// this may be in the superclass?
	private WakeupCriterion[] wakeupEvents;
	private WakeupOr allWakeupEvents;
	protected TransformGroup img_xform;
	protected TransformGroup pbImageXform;
	private int x_last,y_last;
	private Alpha transitionAlpha;
	private PositionInterpolator posmov;
	private ScaleInterpolator scalemov;


	PanZoomMgr(Canvas3D canvas, BranchGroup root, Bounds bounds, 
			TransformGroup transitionXform1, TransformGroup transitionXform2, TransformGroup img_xform, TransformGroup pbImageXform) {
		super(canvas,root,bounds);
		this.img_xform = img_xform;
		this.pbImageXform = pbImageXform;
		this.canvas = canvas;
		
		transitionAlpha = new Alpha(1,300);

		posmov = new PositionInterpolator(transitionAlpha,transitionXform1);
		posmov.setSchedulingBounds(new BoundingSphere(new Point3d(0.0,0.0,0.0),1.0));
		root.addChild(posmov);
		
		scalemov = new ScaleInterpolator(transitionAlpha,transitionXform2);
		scalemov.setSchedulingBounds(new BoundingSphere(new Point3d(0.0,0.0,0.0),1.0));
		root.addChild(scalemov);
	}
	
	public void initialize(){
		wakeupEvents = new WakeupCriterion[4];
		wakeupEvents[0] = new WakeupOnAWTEvent(MouseEvent.MOUSE_MOVED);
		wakeupEvents[1] = new WakeupOnAWTEvent(MouseEvent.MOUSE_DRAGGED);
		wakeupEvents[2] = new WakeupOnAWTEvent(KeyEvent.KEY_PRESSED);
		wakeupEvents[3] = new WakeupOnAWTEvent(KeyEvent.KEY_RELEASED);
		allWakeupEvents = new WakeupOr(wakeupEvents);
		wakeupOn(allWakeupEvents);
	}		
	
	@SuppressWarnings("unchecked")
	public void processStimulus(Enumeration criteria){
		while (criteria.hasMoreElements())
		{
			WakeupCriterion wakeup=(WakeupCriterion) criteria.nextElement();
			if (wakeup instanceof WakeupOnAWTEvent) {
				WakeupOnAWTEvent event = (WakeupOnAWTEvent)wakeup;
				AWTEvent[] events = event.getAWTEvent();
				
				for (int i = 0; i < events.length; ++i) {
					if (events[i] instanceof KeyEvent) {
						KeyEvent keyEvent = (KeyEvent)events[i];
						if (keyEvent.getID() == KeyEvent.KEY_PRESSED)
							if (!processKeyPressed(keyEvent)) {
								wakeupOn(allWakeupEvents);
								processStimulusComplete();
								return;
							}
						if (keyEvent.getID() == KeyEvent.KEY_RELEASED)
							if (!processKeyReleased(keyEvent)) {
								wakeupOn(allWakeupEvents);
								processStimulusComplete();
								return;
							}
						continue;
					}
					
					MouseEvent mev = (MouseEvent)events[i];

					if (mev.getID() == MouseEvent.MOUSE_DRAGGED) {
						// Drag the active keypoint or the image
						if (!processMouseDragged(mev)) {
							wakeupOn(allWakeupEvents);
							processStimulusComplete();
							return;
						}
					} else if (mev.getID() == MouseEvent.MOUSE_MOVED) {
						if (!processMouseMoved(mev)) {
							wakeupOn(allWakeupEvents);
							processStimulusComplete();
							return;
						}
					}
				}
			}
		}
		
		wakeupOn(allWakeupEvents);
		processStimulusComplete();
	}

	public void processStimulusComplete() {}
	@Override
	public void updateScene(int x, int y) {
		System.out.println("Update scene called...");
	}
	
	protected abstract boolean processKeyPressed(KeyEvent ev);
	protected abstract boolean processKeyReleased(KeyEvent ev);
	protected boolean processMouseMoved(MouseEvent mev) {
		x_last = mev.getX();
		y_last = mev.getY();
//		System.out.println("x_last="+x_last+" y_last="+y_last);

		return true;
	}
	
	protected boolean processMouseDragged(MouseEvent mev) {		
		int x=mev.getX();
		int y=mev.getY();
		// using getButton doesn't work on Windows...
		boolean button3Down = (mev.getModifiers() & MouseEvent.BUTTON3_MASK)!=0;
							
		// Move or zoom the image
		
		Transform3D transformDelta = new Transform3D();

		if (button3Down) {
			double z_factor=0.002;
			double z_diff = (y-y_last)*z_factor;
			transformDelta.setScale(1 + z_diff);
		} else {
			double factor=0.002;
			Vector3d translation = new Vector3d();
			translation.x = (x-x_last)*factor; 
			translation.y = -(y-y_last)*factor;

			transformDelta.set(translation);
		}

		Transform3D currXform = new Transform3D();
		img_xform.getTransform(currXform);
		currXform.mul(transformDelta, currXform);
		img_xform.setTransform(currXform);
		pbImageXform.setTransform(currXform);
	
		x_last = x;
		y_last = y;
		return true;
	}
	
	public Point2d curImageCoords() {
		return screenToImagePlane(x_last, y_last);
	}
	
	public Point2d screenToImagePlane(int x, int y) {
		
		Transform3D currXform = new Transform3D();
		img_xform.getTransform(currXform);

		Point3d lwWorld = new Point3d(-1,-1,0);
		currXform.transform(lwWorld);
		Point2d lwScreen=get3DTo2DPoint(lwWorld);
		
		Point3d upWorld = new Point3d(1,1,0);
		currXform.transform(upWorld);
		Point2d upScreen=get3DTo2DPoint(upWorld);		
		
		double fracX = (x-lwScreen.x)/((float)(upScreen.x-lwScreen.x));
		double fracY = (y-lwScreen.y)/((float)(upScreen.y-lwScreen.y));
		
		return new Point2d(fracX,fracY);
	}
	
	private Point2d get3DTo2DPoint(Point3d point3d) {
		Transform3D temp = new Transform3D();
		canvas.getVworldToImagePlate(temp);
		temp.transform(point3d);
		Point2d point2d =new Point2d();
		canvas.getPixelLocationFromImagePlate(point3d,point2d);
		return point2d;
	}
	
	// scroll and zoom so the active keypoints are in full view
	public void resetView(BoundingBox b, boolean doTransition, boolean isImg) {
		Transform3D xformTo = new Transform3D();
		Point3d lower = new Point3d(0.2,0.2,1);
		Point3d upper = new Point3d(0.8,0.8,1);
		if (b!=null) {
			b.getLower(lower);
			b.getUpper(upper);		
		}
		Vector3d lv = new Vector3d(1-(upper.x+lower.x),1-(upper.y+lower.y),0);
		Transform3D t3d = new Transform3D();
		t3d.setTranslation(lv);

		double dist = lower.distance(upper);

		if (dist>0)
			xformTo.setScale(0.7/dist);
	
		xformTo.mul(t3d);
		
		Transform3D xformFrom = new Transform3D();
		img_xform.getTransform(xformFrom);

		Vector3d tFrom = new Vector3d();
		xformFrom.get(tFrom);
		double scaleFrom = xformFrom.getScale();
		Vector3d tTo = new Vector3d();
		xformTo.get(tTo);
		double scaleTo = xformTo.getScale();
		
		double scaleRatio = scaleFrom/scaleTo;
		Vector3d axisOfTranslation = new Vector3d(tFrom);
		tTo.scale(-scaleRatio);
		axisOfTranslation.add(tTo);
				
		double len = axisOfTranslation.length();
		(isImg ? img_xform : pbImageXform).setTransform(xformTo);
		
		if (len>0 || scaleRatio!=1) {
			if (len>0)
				axisOfTranslation.normalize();
			else
				axisOfTranslation.x=1;
			
			posmov.getTarget().setTransform(new Transform3D());
			scalemov.getTarget().setTransform(new Transform3D());	
						
			Transform3D positionAnimationTransform = new Transform3D();
			positionAnimationTransform.rotZ(Math.atan2(axisOfTranslation.y, axisOfTranslation.x));
			
			posmov.setTransformAxis(positionAnimationTransform);	
			posmov.setStartPosition((float)len);
			posmov.setEndPosition(0);
			
			scalemov.setMinimumScale((float)(scaleRatio));
			scalemov.setMaximumScale(1);			

			if (doTransition)
				transitionAlpha.setStartTime(System.currentTimeMillis());
		}				

	}

}
