package core;

import java.util.Vector;

import javax.vecmath.Matrix3d;
import javax.vecmath.Matrix4d;
import javax.vecmath.Point3d;
import javax.vecmath.Vector4d;

import core.Human3DPose.Joint;
import core.Human3DPose.Segment;
import core.Human3DPose.T;

public class PoseOptimizer {
	Vector<Joint> joints;
	Vector<Segment> segments;
	Segment maxSeg;
	private double orig_x[];
	private double orig_y[];
	double dx[];
	double dy[];
	private boolean doingIncrementalCorrections=false;

	private double getElasticity(Joint joint) {
		double el;
		int label=joint.k.getLabel();
		if (label==T.L_Eye || label== T.R_Eye || label==T.L_Ear || label==T.R_Ear || label==T.Nose)
			el=100;
		else if (label==T.L_Knee || label== T.R_Knee || label==T.L_Hip || label==T.R_Hip || label==T.L_Elbow || label==T.R_Elbow)
			el=0.4;
		else
			el=0.8;
		
		if (joint.k.getHidden()) el = el/2;
		return el;
	}

	public void setMaxSeg(Segment maxSeg) { this.maxSeg = maxSeg; }


	public PoseOptimizer(Vector<Joint> joints, Vector<Segment> segments, int numKeypoints) {
		this.joints = joints;
		this.segments = segments;
		orig_x = new double[numKeypoints];
		orig_y = new double[numKeypoints];
		dx = new double[numKeypoints];
		dy = new double[numKeypoints];

	}

	private void taperOffsets() {
		// Taper the offsets so they are not too far from expected center
		double k=0.01;
		for (int j=0; j<joints.size(); ++j) {
			if (dx[j]!=0 && dy[j]!=0) {
				Joint joint=joints.get(j);
				if (!joint.isActive()) continue;
				
				double oldDist = dist(orig_x[j]-joint.x(), orig_y[j]-joint.y());
				double newDist = dist(orig_x[j]-(joint.x()+dx[j]), orig_y[j]-(joint.y()+dy[j]));
				if (oldDist > newDist)
					continue;	// getting closer to orig_xy - great

				double el = getElasticity(joint);

				// want scale to be 0..1 and proportional to deltaDist and inv. prop to oldDist
				double d = Math.pow(k, el) / (1-el) * (Math.pow(k+newDist, 1-el) - Math.pow(k+oldDist, 1-el));
				if (d<=0 || d>=1)
					System.out.println("Warning! d="+d);

				dx[j]*=d;
				dy[j]*=d;
			}
		}		
	}

	private double dist(double x, double y) { return Math.sqrt(x*x+y*y); }

	public void endIncrementalCorrections() {
		doingIncrementalCorrections=false;
	}
	/*
	 	public void fixTorso() {
		Segment shoulderLine = segments.get(Human3DPose.SegmentType.ShoulderLine.ordinal());
		Segment hipLine = segments.get(Human3DPose.SegmentType.HipLine.ordinal());
		Segment spine = segments.get(Human3DPose.SegmentType.Spine.ordinal());
		if (!shoulderLine.isActive() || !hipLine.isActive())
			return;
		double sx = shoulderLine.k.x() - shoulderLine.k_parent.x();
		double sy = shoulderLine.k.y() - shoulderLine.k_parent.y();
		double sz = shoulderLine.k.z() - shoulderLine.k_parent.z();

		double hx = hipLine.k_parent.x() - hipLine.k.x();
		double hy = hipLine.k_parent.y() - hipLine.k.y();
		double hz = hipLine.k_parent.z() - hipLine.k.z();
		
		double d_sl_x=0, d_sl_y=0, d_sr_x=0, d_sr_y=0;
		double d_hl_x=0, d_hl_y=0, d_hr_x=0, d_hr_y=0;
		double old_sl_x=shoulderLine.k.x(),       old_sl_y=shoulderLine.k.y();
		double old_sr_x=shoulderLine.k_parent.x(),old_sr_y=shoulderLine.k_parent.y();
		double old_hr_x=hipLine.     k.x(),       old_hr_y=hipLine.k.y();
		double old_hl_x=hipLine.     k_parent.x(),old_hl_y=hipLine.k_parent.y();
				
		
		if (sx*hx < 0) {
			System.out.println("Flipped shoulders or hips!");
			return;			
		}
		double sign=(sx<0) ? -1 : 1;

		double sxAngle = Math.atan2(sz*sign, sx*sign);
		double hxAngle = Math.atan2(hz*sign, hx*sign);
		if (sxAngle*hxAngle < 0) {
			System.out.println("Twisted depth markers!");
			return;
		}
		Segment shrinkSeg=null, growSeg=null;
		double shrinkStep,growStep;
		if (Math.abs(sxAngle)<Math.abs(hxAngle)) {
			d_sl_x-=Math.abs(sz)*sign*0.01;
			d_sr_x+=Math.abs(sz)*sign*0.01;
			d_hr_x-=Math.abs(hz)*sign*0.01;
			d_hl_x+=Math.abs(hz)*sign*0.01;
			shrinkSeg = shoulderLine; shrinkStep = -Math.abs(sz);
			growSeg = hipLine;        growStep = Math.abs(hz);
			
		} else {
			d_sl_x+=Math.abs(sz)*sign*0.01;
			d_sr_x-=Math.abs(sz)*sign*0.01;
			d_hr_x+=Math.abs(hz)*sign*0.01;
			d_hl_x-=Math.abs(hz)*sign*0.01;
			shrinkSeg = hipLine;      shrinkStep = Math.abs(hz);
			growSeg = shoulderLine;	  growStep = -Math.abs(sz);
		}

		shrinkSeg.k.       k.moveTo(shrinkSeg.k       .x()+shrinkStep*sign*0.01, shrinkSeg.k         .y());
		shrinkSeg.k_parent.k.moveTo(shrinkSeg.k_parent.x()-shrinkStep*sign*0.01, shrinkSeg.k_parent.k.y());

		growSeg  .k.       k.moveTo(growSeg  .k       .x()-growStep  *sign*0.01, growSeg  .k         .y());
		growSeg  .k_parent.k.moveTo(growSeg  .k_parent.x()+growStep  *sign*0.01, growSeg  .k_parent.k.y());

		double syAngle = Math.atan2(sy, Math.sqrt(sx*sx+sz*sz));
		double hyAngle = Math.atan2(hy, Math.sqrt(hx*hx+hz*hz));
		double step = Math.abs(sy-hy)*Math.abs(syAngle-hyAngle);

		if (Math.abs(syAngle)>Math.abs(hyAngle)) {
			shrinkSeg = shoulderLine; 
			growSeg = hipLine;        

			sign = (shoulderLine.k.y() < shoulderLine.k_parent.k.y()) ? 1 : -1;
			d_sl_y+=step*sign*0.1;
			d_sr_y-=step*sign*0.1;
			
			sign = (hipLine.k.y() < hipLine.k_parent.k.y()) ? 1 : -1;
			if (hyAngle*syAngle<0) sign=-sign;
			d_hr_y-=step*sign*0.1;		
			d_hl_y+=step*sign*0.1;			
		} else {
			shrinkSeg = hipLine;      
			growSeg = shoulderLine;	  

			sign = (hipLine.k.y() < hipLine.k_parent.k.y()) ? 1 : -1;
			d_hr_y+=step*sign*0.1;		
			d_hl_y-=step*sign*0.1;

			sign = (shoulderLine.k.y() < shoulderLine.k_parent.k.y()) ? 1 : -1;
			if (hyAngle*syAngle<0) sign=-sign;
			d_sl_y-=step*sign*0.1;
			d_sr_y+=step*sign*0.1;
		}

		
		sign = (shrinkSeg.k.y() < shrinkSeg.k_parent.k.y()) ? 1 : -1;
		shrinkSeg.k.       k.moveTo(shrinkSeg.k       .x(), shrinkSeg.k         .y()+step*sign*0.1);
		shrinkSeg.k_parent.k.moveTo(shrinkSeg.k_parent.x(), shrinkSeg.k_parent.k.y()-step*sign*0.1);

		sign = (growSeg.k.y() < growSeg.k_parent.k.y()) ? 1 : -1;
		if (hyAngle*syAngle<0) sign=-sign;	// when the two angles are the opposite, both shrink
		growSeg  .k.       k.moveTo(growSeg  .k       .x(), growSeg  .k         .y()-step*sign*0.1);
		growSeg  .k_parent.k.moveTo(growSeg  .k_parent.x(), growSeg  .k_parent.k.y()+step*sign*0.1);
		
		// get angle between shoulderline and spine
		Point3d sh = new Point3d();
		Point3d sp = new Point3d();
		double avgLen = (shoulderLine.getDirection(sh) + spine.getDirection(sp))/2;

		
		double spineAngle = Math.acos(sh.x*sp.x + sh.y*sp.y + sh.z*sp.z);
		sign=((spineAngle < Math.PI/2) ? 1 : -1);
		step = Math.abs(spineAngle - Math.PI/2) * avgLen *0.1;
		
		sh.scale(step*sign);
		
		shoulderLine.k.       k.moveTo(shoulderLine.k       .x()+sh.x, shoulderLine.k         .y()+sh.y);
		shoulderLine.k_parent.k.moveTo(shoulderLine.k_parent.x()+sh.x, shoulderLine.k_parent.k.y()+sh.y);
		hipLine.     k.       k.moveTo(hipLine.     k       .x()-sh.x, hipLine.     k         .y()-sh.y);
		hipLine.     k_parent.k.moveTo(hipLine.     k_parent.x()-sh.x, hipLine.     k_parent.k.y()-sh.y);
		
		d_sl_x+=sh.x;
		d_sl_y+=sh.y;
		d_sr_x+=sh.x;
		d_sr_y+=sh.y;
		d_hl_x-=sh.x;
		d_hl_y-=sh.y;
		d_hr_x-=sh.x;
		d_hr_y-=sh.y;
		
		double new_sl_x=shoulderLine.k.x(),       new_sl_y=shoulderLine.k.y();
		double new_sr_x=shoulderLine.k_parent.x(),new_sr_y=shoulderLine.k_parent.y();
		double new_hr_x=hipLine.     k.x(),       new_hr_y=hipLine.k.y();
		double new_hl_x=hipLine.     k_parent.x(),new_hl_y=hipLine.k_parent.y();
		double q=Math.max(Math.abs(old_sl_x+d_sl_x-new_sl_x), Math.abs(old_sl_y+d_sl_y-new_sl_y));
		q = Math.max(q, Math.abs(old_sr_x+d_sr_x-new_sr_x));
		q = Math.max(q, Math.abs(old_sr_y+d_sr_y-new_sr_y));
		q = Math.max(q, Math.abs(old_hl_x+d_hl_x-new_hl_x));
		q = Math.max(q, Math.abs(old_hl_y+d_hl_y-new_hl_y));
		q = Math.max(q, Math.abs(old_hr_x+d_hr_x-new_hr_x));
		q = Math.max(q, Math.abs(old_hr_y+d_hr_y-new_hr_y));
		
		System.out.print(Math.abs(old_sl_x+d_sl_x-new_sl_x)+" "+Math.abs(old_sl_y+d_sl_y-new_sl_y)+" ");
		System.out.print(Math.abs(old_sr_x+d_sr_x-new_sr_x)+" "+Math.abs(old_sr_y+d_sr_y-new_sr_y)+" ");
		System.out.print(Math.abs(old_hl_x+d_hl_x-new_hl_x)+" "+Math.abs(old_hl_y+d_hl_y-new_hl_y)+" ");
		System.out.println(Math.abs(old_hr_x+d_hr_x-new_hr_x)+" "+Math.abs(old_hr_y+d_hr_y-new_hr_y));
		System.out.println("Q="+q);
	}
	 
	// Moves the hips and shoulders in a direction towards equilateral trapezoid
	public void fixTorso2() {
		Segment shoulderLine = segments.get(Human3DPose.SegmentType.ShoulderLine.ordinal());
		Segment hipLine = segments.get(Human3DPose.SegmentType.HipLine.ordinal());
		Segment spine = segments.get(Human3DPose.SegmentType.Spine.ordinal());
		if (!shoulderLine.isActive() || !hipLine.isActive())
			return;
		double sx = shoulderLine.k.x() - shoulderLine.k_parent.x();
		double sy = shoulderLine.k.y() - shoulderLine.k_parent.y();
		double sz = shoulderLine.k.z() - shoulderLine.k_parent.z();

		double hx = hipLine.k_parent.x() - hipLine.k.x();
		double hy = hipLine.k_parent.y() - hipLine.k.y();
		double hz = hipLine.k_parent.z() - hipLine.k.z();
		
		double d_sl_x=0, d_sl_y=0, d_sr_x=0, d_sr_y=0;
		double d_hl_x=0, d_hl_y=0, d_hr_x=0, d_hr_y=0;
						
		if (sx*hx < 0) {
			System.out.println("Flipped shoulders or hips!");
			return;			
		}
		double sign=(sx<0) ? -1 : 1;

		double sxAngle = Math.atan2(sz*sign, sx*sign);
		double hxAngle = Math.atan2(hz*sign, hx*sign);
		if (sxAngle*hxAngle < 0) {
			System.out.println("Twisted depth markers!");
			return;
		}
		if (Math.abs(sxAngle)<Math.abs(hxAngle)) {
			d_sl_x-=Math.abs(sz)*sign*0.01;
			d_sr_x+=Math.abs(sz)*sign*0.01;
			d_hr_x-=Math.abs(hz)*sign*0.01;
			d_hl_x+=Math.abs(hz)*sign*0.01;
		} else {
			d_sl_x+=Math.abs(sz)*sign*0.01;
			d_sr_x-=Math.abs(sz)*sign*0.01;
			d_hr_x+=Math.abs(hz)*sign*0.01;
			d_hl_x-=Math.abs(hz)*sign*0.01;
		}


		double syAngle = Math.atan2(sy, Math.sqrt(sx*sx+sz*sz));
		double hyAngle = Math.atan2(hy, Math.sqrt(hx*hx+hz*hz));
		double step = Math.abs(sy-hy)*Math.abs(syAngle-hyAngle);

		if (Math.abs(syAngle)>Math.abs(hyAngle)) {
			sign = (shoulderLine.k.y() < shoulderLine.k_parent.k.y()) ? 1 : -1;
			d_sl_y+=step*sign*0.1;
			d_sr_y-=step*sign*0.1;
			
			sign = (hipLine.k.y() < hipLine.k_parent.k.y()) ? 1 : -1;
			if (hyAngle*syAngle<0) sign=-sign;
			d_hr_y-=step*sign*0.1;		
			d_hl_y+=step*sign*0.1;			
		} else {
			sign = (hipLine.k.y() < hipLine.k_parent.k.y()) ? 1 : -1;
			d_hr_y+=step*sign*0.1;		
			d_hl_y-=step*sign*0.1;

			sign = (shoulderLine.k.y() < shoulderLine.k_parent.k.y()) ? 1 : -1;
			if (hyAngle*syAngle<0) sign=-sign;
			d_sl_y-=step*sign*0.1;
			d_sr_y+=step*sign*0.1;
		}
		
		// get angle between shoulderline and spine
		Point3d sh = new Point3d();
		Point3d sp = new Point3d();
		double avgLen = (shoulderLine.getDirection(sh) + spine.getDirection(sp))/2;

		
		double spineAngle = Math.acos(sh.x*sp.x + sh.y*sp.y + sh.z*sp.z);
		sign=((spineAngle < Math.PI/2) ? 1 : -1);
		step = Math.abs(spineAngle - Math.PI/2) * avgLen *0.1;
		
		sh.scale(step*sign*0);

		d_sl_x+=sh.x;
		d_sl_y+=sh.y;
		d_sr_x+=sh.x;
		d_sr_y+=sh.y;
		d_hl_x-=sh.x;
		d_hl_y-=sh.y;
		d_hr_x-=sh.x;
		d_hr_y-=sh.y;
		
		double mean_x = (d_sl_x+d_sr_x+d_hl_x+d_hr_x)/4;
		double mean_y = (d_sl_y+d_sr_y+d_hl_y+d_hr_y)/4;
		
		shoulderLine.k.       k.moveTo(shoulderLine.k       .x()+d_sl_x-mean_x, shoulderLine.k         .y()+d_sl_y-mean_y);
		shoulderLine.k_parent.k.moveTo(shoulderLine.k_parent.x()+d_sr_x-mean_x, shoulderLine.k_parent.k.y()+d_sr_y-mean_y);
		hipLine.     k.       k.moveTo(hipLine.     k       .x()+d_hr_x-mean_x, hipLine.     k         .y()+d_hr_y-mean_y);
		hipLine.     k_parent.k.moveTo(hipLine.     k_parent.x()+d_hl_x-mean_x, hipLine.     k_parent.k.y()+d_hl_y-mean_y);
	}
	*/
	public void fixTorso() {
		Segment shoulderLine = segments.get(Human3DPose.SegmentType.ShoulderLine.ordinal());
		Segment hipLine = segments.get(Human3DPose.SegmentType.HipLine.ordinal());
		Keypoint slk = shoulderLine.k.k;
		Keypoint srk = shoulderLine.k_parent.k;
		Keypoint hlk = hipLine.k_parent.k;
		Keypoint hrk = hipLine.k.k;
		
		Point3d sl = slk.getCoordinates();
		Point3d sr = srk.getCoordinates();
		Point3d hl = hlk.getCoordinates();
		Point3d hr = hrk.getCoordinates();
		
		Point3d sl_target = new Point3d(sl);
		Point3d sr_target = new Point3d(sr);
		Point3d hl_target = new Point3d(hl);
		Point3d hr_target = new Point3d(hr);
		
		isoscalesTrapezoid(hl,hr,sl,sr_target);
		isoscalesTrapezoid(hr,hl,sr,sl_target);
		isoscalesTrapezoid(sl,sr,hl,hr_target);
		isoscalesTrapezoid(sr,sl,hr,hl_target);	
		
		/*
		sl_target.sub(sl);
		sr_target.sub(sr);
		hl_target.sub(hl);
		hr_target.sub(hr);
		
		// Center the data
		Point3d mean=new Point3d(sl_target);
		mean.add(sr_target);
		mean.add(hl_target);
		mean.add(hr_target);
		mean.scale(0.25);
		sl_target.sub(mean);
		sr_target.sub(mean);
		hl_target.sub(mean);
		hr_target.sub(mean);
		
		double step=0.01;
		sl_target.scale(step);
		sr_target.scale(step);
		hl_target.scale(step);
		hr_target.scale(step);
		

		double mean_x = (sl_target.x+sr_target.x+hl_target.x+hr_target.x)/4;
		double mean_y = (sl_target.y+sr_target.y+hl_target.y+hr_target.y)/4;

		slk.move(sl_target.x-mean_x, sl_target.y-mean_y);
		srk.move(sr_target.x-mean_x, sr_target.y-mean_y);
		hlk.move(hl_target.x-mean_x, hl_target.y-mean_y);
		hrk.move(hr_target.x-mean_x, hr_target.y-mean_y);
		*/
		
		double[] x1 = {hl.x, hr.x, sl.x, sr.x};
		double[] y1 = {hl.y, hr.y, sl.y, sr.y};
		double[] x2 = {hl_target.x, hr_target.x, sl_target.x, sr_target.x};
		double[] y2 = {hl_target.y, hr_target.y, sl_target.y, sr_target.y};
	
		Matrix3d simXForm = extractSimilarityTransform(x1,y1,x2,y2);	
		sl.z=1; simXForm.transform(sl);
		sr.z=1; simXForm.transform(sr);
		hl.z=1; simXForm.transform(hl);
		hr.z=1; simXForm.transform(hr);
		sl_target.sub(sl);
		sr_target.sub(sr);
		hl_target.sub(hl);
		hr_target.sub(hr);
		
		double step=0.01;
		sl_target.scale(step);
		sr_target.scale(step);
		hl_target.scale(step);
		hr_target.scale(step);

		slk.move(sl_target.x, sl_target.y);
		srk.move(sr_target.x, sr_target.y);
		hlk.move(hl_target.x, hl_target.y);
		hrk.move(hr_target.x, hr_target.y);
	}
	
	// Returns in D the coordinates that make isosceles trapezoid with bases AB and CD
	void isoscalesTrapezoid(Point3d A, Point3d B, Point3d C, Point3d D) {
		double ax=B.x-A.x, ay=B.y-A.y, az=B.z-A.z;
		double hx=(B.x+A.x)/2, hy=(B.y+A.y)/2, hz=(B.z+A.z)/2;
		double denom = ax*ax + ay*ay + az*az;
		if (denom==0)
			denom=0.000000000001;
		double lambda=2*((hx-C.x)*ax+(hy-C.y)*ay+(hz-C.z)*az) / denom;
		D.set(C.x+lambda*ax, C.y+lambda*ay, C.z+lambda*az);
	}
	
	// Decomposes a mapping of points [x1 y1] into points [x2 y2] into
	// a similarity transformation (translation, rotation, uniform scale) and error term for each point
	Matrix3d extractSimilarityTransform(double x1[], double y1[], double x2[], double y2[]) {
		double x1sum=0;
		double y1sum=0;
		double x1sqrd_plus_y1sqrd=0;
		double x1x2_plus_y1y2=0;
		double x2sum=0;
		double y2sum=0;
		double x2y1_minus_x1y2=0;
		double N=x1.length;
		for (int i=0; i<N; ++i) {
			x1sum+=x1[i];
			y1sum+=y1[i];
			x2sum+=x2[i];
			y2sum+=y2[i];
			x1sqrd_plus_y1sqrd+=x1[i]*x1[i]+y1[i]*y1[i];
			x1x2_plus_y1y2+=x1[i]*x2[i] + y1[i]*y2[i];
			x2y1_minus_x1y2+=(x2[i]*y1[i] - x1[i]*y2[i]);
		}
		Matrix4d AAt = new Matrix4d(x1sqrd_plus_y1sqrd, 0,  x1sum,  y1sum,
				                    0,  x1sqrd_plus_y1sqrd, y1sum, -x1sum,
				                    x1sum,     y1sum,    N,      0,
				                    y1sum,    -x1sum,    0,      N);
		AAt.invert();
		Vector4d r = new Vector4d(x1x2_plus_y1y2, x2y1_minus_x1y2, x2sum, y2sum);
		AAt.transform(r);
		double tx = r.z, ty=r.w;
		// r.x = scale*cos(a)  r.y = scale*sin(a)
//		double scale = Math.sqrt(r.x*r.x + r.y*r.y);
//		double angle = Math.acos(r.x/scale);
		return new Matrix3d(r.x, r.y, tx,
				           -r.y, r.x, ty,
				            0,  0,  1);
	}
		
	public void doIncrementalCorrections(boolean justMakeMaxSegBigger, boolean faster) {
		int J = joints.size();
		double step = (faster ? 0.4 : 0.1);

//		if (justMakeMaxSegBigger)
//			System.out.println("justMakeMaxSegBigger");

		if (!doingIncrementalCorrections) {
//			System.out.println("Reset incremental");
			for (int j=0; j<J; ++j) {
				orig_x[j] = joints.get(j).x();
				orig_y[j] = joints.get(j).y();
			}
			doingIncrementalCorrections=true;
		}

		for (int j=0; j<J; ++j) {
			dx[j]=0;
			dy[j]=0;
		}
		
		double sumDist=0;
		int numActive = 0;
		for (Segment s : segments)
			if (s.isActive()) {
				double relX = s.k.x()-s.k_parent.x();
				double relY = s.k.y()-s.k_parent.y();
				sumDist += Math.sqrt(relX*relX + relY*relY);				
				numActive++;
			}
		if (numActive==0)
			return;
		double avgDist = sumDist/numActive;
//		System.out.println("AvgDist="+avgDist);
				

		// When you adjust the hip or the shoulder line, adjust the other too so that we don't get a twist
		Segment maxSeg2=null;
		if (justMakeMaxSegBigger && maxSeg!=null) {
			if (maxSeg.type()==Human3DPose.SegmentType.ShoulderLine)
				maxSeg2=segments.get(Human3DPose.SegmentType.HipLine.ordinal());
			else if (maxSeg.type()==Human3DPose.SegmentType.HipLine)
				maxSeg2=segments.get(Human3DPose.SegmentType.ShoulderLine.ordinal());
		}

		for (Segment s : segments)
			if (s.isActive() && (s.k.getZOrder()==Keypoint.ZOrder.Equal || 
					(s==maxSeg && (justMakeMaxSegBigger || (s.k.getZOrder()!=Keypoint.ZOrder.Equal))))) {
				boolean isMaxSeg = ((s==maxSeg || s==maxSeg2) && (justMakeMaxSegBigger || (s.k.getZOrder()!=Keypoint.ZOrder.Equal)));
				if (justMakeMaxSegBigger && !isMaxSeg)
					continue;

				// This is a segment that ought to be parallel to the image plane, but it is xy length is too small - enlarge it
				double relX = s.k.x()-s.k_parent.x();
				double relY = s.k.y()-s.k_parent.y();
				double relZ = Math.abs(s.k.z() - s.k_parent.z());
				double length2DSqrd = relX*relX + relY*relY;
				double length2D = Math.sqrt(length2DSqrd);
				// each parent needs to be pushed by desiredCorrection distance away from the other to make relZ==0
				// Or if s==maxSeg, our segment is the shortest one and is considered parallel to the image plane, yet it is not marked with Equal. Shrink it
				double desiredCorrection = (isMaxSeg) ? (justMakeMaxSegBigger ? -step*2*avgDist : -step*avgDist) : (Math.sqrt(length2DSqrd + relZ*relZ)-length2D)/2;
				double dx1 = relX/length2D*desiredCorrection;
				double dy1 = relY/length2D*desiredCorrection;

				int i;
				i = s.k       .k.getLabel(); dx[i]+=dx1; dy[i]+=dy1;
				i = s.k_parent.k.getLabel(); dx[i]-=dx1; dy[i]-=dy1;

				if (s.k.k.getLabel()==T.M_Hip) {
					// We are the spine. Move the shoulders and hips
					i = T.L_Hip; dx[i]+=dx1; dy[i]+=dy1;
					i = T.R_Hip; dx[i]+=dx1; dy[i]+=dy1;
					i = T.L_Shoulder; dx[i]-=dx1; dy[i]-=dy1;
					i = T.R_Shoulder; dx[i]-=dx1; dy[i]-=dy1;
				} else if (s.k.k.getLabel()==T.L_Hip || s.k.k.getLabel()==T.R_Hip) {
					i = T.M_Hip; dx[i]+=dx1/2; dy[i]+=dy1/2;
				}
				if (s.k_parent.k.getLabel()==T.L_Hip || s.k_parent.k.getLabel()==T.R_Hip) {
					i = T.M_Hip; dx[i]-=dx1/2; dy[i]-=dy1/2;
				}
			}

		taperOffsets();			
		// find the scale that, when moving each joint along that scale, the sum of squared errors is minimized

		double num=0, denom=0;
		for (int j=0; j<J; ++j) {
			Joint joint=joints.get(j);
			if (joint.isActive()) {
				num+=  (joint.x()          * (joint.x() + dx[j])) + (joint.y()          * (joint.y() + dy[j]));
				denom+=(joint.x() + dx[j]) * (joint.x() + dx[j])  + (joint.y() + dy[j]) * (joint.y() + dy[j]);
			}
		}
		double scale = num/denom;
		/*
		double maxD = 0;
		for (int j=0; j<J; ++j) {
			maxD = Math.max(maxD, Math.max(dx[j], dy[j]));
		}
		double thresh=10000;
		System.out.println("Max move: "+maxD*scale*thresh);
		if (maxD*scale*thresh<1)
			scale = 1/(thresh*maxD);
		*/
	//	double maxMoveDist = 0;
		for (int j=0; j<J; ++j) {
			Joint joint=joints.get(j);
			if (!joint.isActive())
				continue;

			if (j==T.L_Eye || j==T.R_Eye || j==T.L_Ear || j==T.R_Ear || j==T.Nose) {
				dx[j]=0;
				dy[j]=0;
			} else {
				joint.k.moveTo((joint.x()+dx[j])*scale, (joint.y()+dy[j])*scale);
//				if (maxMoveDist > dx[j]*dx[j]+dy[j]*dy[j])
//					maxMoveDist = dx[j]*dx[j]+dy[j]*dy[j];
				joint.setZ(joint.z()*scale);
			}
		}
//		System.out.println("Max move dist "+maxMoveDist);
	}
}
