package core;

import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;

import javax.imageio.ImageIO;
import javax.media.j3d.Appearance;
import javax.media.j3d.Bounds;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Canvas3D;
import javax.media.j3d.GeometryArray;
import javax.media.j3d.ImageComponent2D;
import javax.media.j3d.QuadArray;
import javax.media.j3d.RenderingAttributes;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Texture;
import javax.media.j3d.TransformGroup;
import javax.swing.JOptionPane;
import javax.vecmath.Point2d;
import javax.vecmath.Point3d;
import javax.vecmath.TexCoord2f;

import com.sun.j3d.utils.image.TextureLoader;


public class SegmentationMgr extends PanZoomMgr {
	private String[] regionNames;
	private int[] regionColors;
	private char[] shortcuts;
	private boolean suppressKeyEvents = false;
	private int segmentationLevel = 0;
	private Superpixel hilightedSuperpixel = null;
	private Set<Border> hilightedOutline = new HashSet<Border>();
	private Set<Border> hilightedRegionOutline = new HashSet<Border>(); // outline of the entire region
	private int[] labelsImage=null;
	private int labelsImgW, labelsImgH;
	private BufferedImage displayImg=null, leftDisplayImg=null;
	private BufferedImage src;
	private ArrayList<Superpixel> superpixels;
	private ArrayList<Integer>[] regionLabels;	// stores them here if not enabled. Updates once enabled
	private String pbFileName;				// stores it if not enabled. Updates once enabled
	private boolean isEnabled=false;	// true if the current view is segmentation
	private LinkedList<SuperpixelValuePair> prevLabels = null;
	private boolean keyIsDown=false;
	private int scopeLabel=-2;
	private int currentLabel=-1;
	
	private Shape3D imageNode, leftImageNode;

	private int imgWidth, imgHeight;
	private UIManager uiMgr;

	class ShortPoint {
		short x,y;
		ShortPoint(short x, short y) { this.x = x; this.y = y; }
	}

	class SuperpixelValuePair {
		public SuperpixelValuePair(Superpixel s, int value) {
			this.s = s;
			this.value = value;
		}
		Superpixel s;
		int value;
	}

	class Region {
		private ArrayList<ShortPoint> pixels = new ArrayList<ShortPoint>();
		private int[] compact_pixels;
		
		public void add_pixel(short x, short y) { pixels.add(new ShortPoint(x,y)); }		
		
		public int size() { return compact_pixels.length; }
		public int get_x(int idx) { return compact_pixels[idx] % 16384; }
		public int get_y(int idx) { return compact_pixels[idx] / 16384; }
				
		public void compact() {	// call after all pixels are collected. Compacts the data
			if(compact_pixels!=null)
				return;
						
			compact_pixels = new int[pixels.size()];
			for (int i=0; i<pixels.size(); ++i)
				compact_pixels[i] = pixels.get(i).y*16384 + pixels.get(i).x;
			
			pixels.clear();
		}
	}
	
	class Border extends Region {
		private Superpixel p1, p2;
		private int pb;
		private boolean boundaryDrawn;
		
		Border(Superpixel p1, Superpixel p2, int pb) {
			this.p1 = p1;
			this.p2 = p2;
			this.pb = pb;
			boundaryDrawn = false;
			assert(p1!=p2 && pb>0 && pb<256);
		}
		
		public int pb() { return pb; }
		public Superpixel other(Superpixel p) { return p==p1 ? p2 : p1; }
		
		public void clear_boundary() {
			boundaryDrawn = false;
			for (int i=0; i<size(); ++i)
				displayImg.setRGB(get_x(i),get_y(i),src.getRGB(get_x(i),get_y(i)));
		}
		public void update_boundary() {
			if (should_be_drawn()) {
				if (boundaryDrawn) return;
				for (int i=0; i<size(); ++i) 
					displayImg.setRGB(get_x(i),get_y(i),0xFFFFFF);
				boundaryDrawn = true;
			} else
				if (boundaryDrawn) 
					clear_boundary();
		}
		
		public void set_hilighted(boolean hilighted) {
			if (hilighted) {
				draw(displayImg, 0xFF0000);
			} else {
				if (should_be_drawn()) {
					draw(displayImg, 0xFFFFFF);
				} else {
					// to recover the color force one superpixel to redraw
					assert(p1.label()==p2.label());
					if (p1.label()==-1)
						erase(displayImg);
					else {
						int color = p1.get_color();
						for (int i=0; i<size(); ++i) 
							displayImg.setRGB(get_x(i),get_y(i),p1.marks_pixel(get_x(i),get_y(i)) ? color : src.getRGB(get_x(i),get_y(i)));
					}
				}
			}
		}
		public void draw(BufferedImage img, int color) {
			for (int i=0; i<size(); ++i)
				img.setRGB(get_x(i),get_y(i), color);			
		}
		
		public void erase(BufferedImage img) {
			for (int i=0; i<size(); ++i)
				img.setRGB(get_x(i),get_y(i), src.getRGB(get_x(i),get_y(i)));			
		}

		private boolean should_be_drawn() {
			return (p1.label()!=p2.label()) || (p1.label()==-1 && pb>=segmentationLevel);
		}
	}

	class Superpixel extends Region {
		private ArrayList<Border> borders = new ArrayList<Border>();
		
		Superpixel(int id) {
			this.id = id;
		}
		
		private int label=-1;
		private int id;
		
		public int label() { return label; }
		public int id() { return id; }
		public ArrayList<Border> borders() { return borders; }
		
		void add_neighbor(Superpixel neighbor, int pb, short x, short y) {
			for (Border b : borders)
				if (b.other(this)==neighbor) {
					assert(b.pb() == pb);
					b.add_pixel(x,y);
					return;
				}
			Border b = new Border(this,neighbor,pb);
			b.add_pixel(x,y);
			borders.add(b);
			neighbor.borders.add(b);
		}
		
		void add_border_pixel(Superpixel neighbor, int pb, short x, short y) {
			for (Border b : borders) {
				if (b.other(this)==neighbor && b.pb()==pb) {
					b.add_pixel(x,y);
					return;
				}
			}
			assert(false);
		}
		
		int get_color() {
			assert(label!=-1);
			return regionColors[label];
		}
		
		// collects borders of the region that includes this superpixel (based on the current segmentation threshold)
		private void get_region(Set<Superpixel> region, Set<Border> outline) {
			outline.clear();
			region.clear();
			LinkedList<Superpixel> q=new LinkedList<Superpixel>();
			q.add(this);
			while (!q.isEmpty()) {
				Superpixel s = q.remove();
				if (region.add(s)) {
					for (Border b : s.borders())
						if (b.pb()>=segmentationLevel || (scopeLabel!=-2 && b.other(s).label()!=scopeLabel))
							outline.add(b);
						else
							q.add(b.other(s));
				}
			}
		}
		public void get_region_outline(Set<Border> outline) {
			get_region(new HashSet<Superpixel>(), outline);
		}
		
		// collects just the superpixels of the region that includes this superpixel (based on the current segmentation threshold)
		void get_region_area(Set<Superpixel> region) {
			get_region(region, new HashSet<Border>());
		}
				
		boolean marks_pixel(int x, int y) { return (x % 3)==0 || (y % 3)==0; }
		
		void superpixel_set_label(int label) {
			if (this.label==label)
				return;
			this.label = label;

			for (Border b : borders)
				b.clear_boundary();				
				
			paint_superpixel();
						
			for (Border b : borders)
				b.update_boundary();	
		}
		
		
		void region_set_label(int label) {			
			Set<Superpixel> region = new HashSet<Superpixel>();
			get_region_area(region);

//			LinkedList<SuperpixelValuePair> prevLabels = new LinkedList<SuperpixelValuePair>();
			boolean allMarked = true;
			for (Superpixel s : region) {
				if (s.label()!=label)
					allMarked = false;
					prevLabels.add(new SuperpixelValuePair(s, s.label()));
				}
				
			if (allMarked)
				return;

			
			for (Superpixel s : region) {
				s.label = label;
				for (Border b : s.borders())
					b.clear_boundary();	
			}
				
			for (Superpixel s : region)
				s.paint_superpixel();
			
			for (Superpixel s : region)
				for (Border b : s.borders())
					b.update_boundary();
		}
	
		private void paint_superpixel() {
			for (int i=0; i<size(); ++i) {
				if (marks_pixel(get_x(i),get_y(i)))
					displayImg.setRGB(get_x(i),get_y(i),label==-1 ? src.getRGB(get_x(i),get_y(i)) : regionColors[label]);
			}	
		}
	}
	


	public boolean hasSegments() { 
		if (pbFileName==null) return false;
		if (uiMgr.isReadOnly())
			return Thread.currentThread().getContextClassLoader().getResource(pbFileName)!=null; 

		File f = new File(pbFileName);
		return f.exists();
	}
	public Shape3D getImage() { return imageNode; }
	public int numRegions() { return shortcuts.length; }
	private double getAspectRatio() { return imgWidth / (double)imgHeight; }
	
	SegmentationMgr(Canvas3D canvas, BranchGroup root, Bounds bounds, 
			TransformGroup xform1, TransformGroup xform2, TransformGroup img_xform, TransformGroup pbImageXform,
			UIManager uiMgr, ConfigMgr configMgr) {
		super(canvas,root,bounds,xform1,xform2,img_xform,pbImageXform);
		this.uiMgr = uiMgr;
		this.shortcuts = configMgr.getRegionShortcuts();
		this.regionNames = configMgr.getRegionNames();
		this.regionColors = configMgr.getRegionColors();
		QuadArray plane = new QuadArray(4, GeometryArray.COORDINATES | GeometryArray.TEXTURE_COORDINATE_2);

		TexCoord2f q = new TexCoord2f();
		q.set(0.0f, 0.0f);	plane.setTextureCoordinate(0, 0, q);
		q.set(1.0f, 0.0f);  plane.setTextureCoordinate(0, 1, q);
		q.set(1.0f, 1.0f);	plane.setTextureCoordinate(0, 2, q);
		q.set(0.0f, 1.0f);	plane.setTextureCoordinate(0, 3, q);
		
		RenderingAttributes ra = new RenderingAttributes();
		Appearance ap = new Appearance();
		ap.setRenderingAttributes(ra);
		ap.setCapability(Appearance.ALLOW_TEXTURE_WRITE);
		imageNode = new Shape3D(plane, ap);
		imageNode.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
		imageNode.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
		imageNode.getGeometry().setCapability(QuadArray.ALLOW_COORDINATE_WRITE);
	}
	

	public Point2d unitToPixelCoords(double ux, double uy) {
		double x = (ux + (getAspectRatio()-1.0)/2.0)*imgHeight;
		double y = (1 - uy)*imgHeight;
		return new Point2d(x,y);
	}
	
	private boolean make_labels_image() {
		labelsImage = null;
		Raster bufLabels;
		
		try {
			if (uiMgr.isReadOnly()) {
				bufLabels = ((BufferedImage)ImageIO.read(Thread.currentThread().getContextClassLoader().getResource(pbFileName))).getRaster();
			} else {
				java.io.File file = new java.io.File(pbFileName);
				bufLabels = ((BufferedImage)ImageIO.read(file)).getRaster();
			}
		} catch (IOException e) {
			return false;
		}

		labelsImgW = bufLabels.getWidth();
		labelsImgH = bufLabels.getHeight();
		labelsImage = new int[labelsImgH*labelsImgW];
		
		for (int y=0; y<labelsImgH; ++y)
			for (int x=0; x<labelsImgW; ++x)
				labelsImage[y*labelsImgW + x] = bufLabels.getSample(x,y, 0);
		return true;
	}
	private int label_at(int x, int y) { return labelsImage[y*labelsImgW + x]; }
	
	private boolean is_border(int x, int y) {
		return label_at(x,y)>=16384;
	}
	private Superpixel get_superpixel(int x, int y) {
		if (x<0 || y<0 || x>=labelsImgW || y>=labelsImgH)
			return null;
		int val = label_at(x,y);
		if (val >= 16384) return null;
		return superpixels.get(val-1);
	}
	
	void setImage(String pbFileName, BufferedImage src, Shape3D leftImgNode) {
		hilightedOutline.clear();
		hilightedRegionOutline.clear();
		hilightedSuperpixel = null;
		superpixels = null;
		displayImg = null;
		this.pbFileName = pbFileName;
		this.src = src;
		this.leftImageNode = leftImgNode;

		
		init();
	}

	// create the superpixels and borders for this image
	private void init() {
		if (!isEnabled) return;
		if (!make_labels_image())
			return;

		// get the number of superpixels
		int numSuperpixels=0;
		for (int y=0; y<labelsImgH; ++y)
			for (int x=0; x<labelsImgW; ++x) {
				int val = label_at(x,y);
				if (val<16384 && val>numSuperpixels)
					numSuperpixels=val;
			}
		
		superpixels = new ArrayList<Superpixel>(numSuperpixels);
		for (int i=0; i<numSuperpixels; ++i)
			superpixels.add(new Superpixel(i));
		
		int W = labelsImgW-1, H = labelsImgH-1;
		for (short y=0; y<=H; ++y)
			for (short x=0; x<=W; ++x) {
				int val = label_at(x,y);
				if (val<16384)
					superpixels.get(val-1).add_pixel(x,y);
				else {
					for (int xx=x-1; xx<=x+1; ++xx)
						for (int yy=y-1; yy<=y+1; ++yy) {
							Superpixel s = get_superpixel(xx,yy);
							if (s!=null) s.add_pixel(x,y);	// the border pixels are added to all adjacent superpixels
						}
					int pb = val-16384;
					if ((x==0 || is_border(x-1,y)) && (x==W || is_border(x+1,y))) {
						// ? A ?
						// 0 0 0
						// ? B ?
						Superpixel a = get_superpixel(x,y-1);
						Superpixel b = get_superpixel(x,y+1);
						if (a!=null && b!=null)
							a.add_neighbor(b, pb, x, y);

					} else if ((y==0 || is_border(x,y-1)) && (y==H || is_border(x,y+1))) {
				        // ? 0 ?
				        // A 0 B
				        // ? 0 ?
						Superpixel a = get_superpixel(x-1,y);
						Superpixel b = get_superpixel(x+1,y);
						if (a!=null && b!=null)
							a.add_neighbor(b, pb, x, y);
					}
				}
			}
		
		// Add each border pixel to a border region
		Superpixel[][] patch = new Superpixel[3][3];
		for (short y=0; y<=H; ++y)
			for (short x=0; x<=W; ++x) {
				int val = label_at(x,y);
				if (val>=16384) {
					int pb = val-16384;
					for (int xx=0; xx<=2; ++xx)
						for (int yy=0; yy<=2; ++yy)
							patch[xx][yy] = get_superpixel(x+xx-1, y+yy-1);

					if (patch[0][1]==null && patch[1][0]==null) {
			            // A 0 ?
			            // 0 0 ?
			            // ? ? B
						if (patch[0][0]!=null && patch[2][2]!=null)
							patch[0][0].add_border_pixel(patch[2][2],pb,x,y);
					}
					if (patch[0][1]==null && patch[1][2]==null) {
			            // ? ? B
			            // 0 0 ?
			            // A 0 ?
						if (patch[0][2]!=null && patch[2][0]!=null)
							patch[0][2].add_border_pixel(patch[2][0],pb,x,y);
					}
					if (patch[1][0]==null && patch[2][1]==null) {
			            // ? 0 A
			            // X 0 0
			            // B ? ?
						if (patch[2][0]!=null && patch[0][2]!=null)
							patch[2][0].add_border_pixel(patch[0][2],pb,x,y);
					}
					if (patch[2][1]==null && patch[1][2]==null) {
			            // B X ?
			            // X 0 0
			            // ? 0 A
						if (patch[2][2]!=null && patch[0][0]!=null)
							patch[2][2].add_border_pixel(patch[0][0],pb,x,y);
					}					
				}
			}
		
		for (Superpixel s : superpixels) {
			s.compact();
			for (Border b : s.borders())
				b.compact();
		}
				
		displayImg = new BufferedImage(labelsImgW,labelsImgH, BufferedImage.TYPE_3BYTE_BGR);
		displayImg.setData(src.getData());

		leftDisplayImg = new BufferedImage(labelsImgW,labelsImgH, BufferedImage.TYPE_3BYTE_BGR);
		leftDisplayImg.setData(src.getData());
		

		TextureLoader loader = new TextureLoader(displayImg,"RGB");//,TextureLoader.BY_REFERENCE);
		
		imgWidth = loader.getImage().getWidth();
		imgHeight = loader.getImage().getHeight();
		double aspectRatio = getAspectRatio();

		QuadArray plane = (QuadArray)imageNode.getGeometry();

		Point3d p = new Point3d();
		
		p.set(-aspectRatio, -1,  0);	plane.setCoordinate(0, p);
		p.set(aspectRatio,  -1,  0);	plane.setCoordinate(1, p);
		p.set(aspectRatio,   1,  0);	plane.setCoordinate(2, p);
		p.set(-aspectRatio,  1,  0);	plane.setCoordinate(3, p);

		loader.getTexture().setCapability(Texture.ALLOW_IMAGE_WRITE);
		loader.getTexture().getImage(0).setCapability(ImageComponent2D.ALLOW_IMAGE_WRITE);
		loader.getTexture().getImage(0).setCapability(ImageComponent2D.ALLOW_IMAGE_READ);
		imageNode.getAppearance().setTexture(loader.getTexture());		
	}
	
	void setRegionLabels(ArrayList<Integer>[] labels) {
		regionLabels = labels;
		updateRegionLabels();
	}
	
	private void updateRegionLabels() {
		if (!isEnabled) return;
		if (superpixels.isEmpty() || regionLabels==null) return;
		prevLabels = null;	// clear undo info
		
		for (Superpixel p : superpixels)
			p.superpixel_set_label(-1);
		for (int label=0; label<regionLabels.length; ++label)
			for (int id : regionLabels[label])
				superpixels.get(id-1).superpixel_set_label(label);
		
		for (Superpixel s : superpixels)
			for (Border b : s.borders())
				b.update_boundary();
		
		updateTexture(true);
		regionLabels = null;
	}
	
	@SuppressWarnings("unchecked")
	ArrayList<Integer>[] getRegionLabels() {
		ArrayList<Integer>[] labels = new ArrayList[regionNames.length];
		for (int i=0; i<labels.length; ++i) {
			labels[i] = new ArrayList<Integer>();
			for (Superpixel s : superpixels) {
				if (s.label()!=-1 && s.label() == i)
					labels[i].add(s.id()+1);
			}
		}
		return labels;
	}
	
	@Override
	protected boolean processMouseDragged(MouseEvent mev) {		
		if (suppressKeyEvents) {
			return false;
		}
		return super.processMouseDragged(mev);
	}
		
	@Override
	protected boolean processMouseMoved(MouseEvent mev) {
		if (suppressKeyEvents) {
			return false;
		}
		try {
			super.processMouseMoved(mev);
	
			Point2d coords = curImageCoords();
			Point2d pixCoords = unitToPixelCoords(coords.x,coords.y);
	
			int x=(int)Math.round(pixCoords.x);
			int y=(int)Math.round(pixCoords.y);
			if (x<0 || y<0 || x>=imgWidth || y>=imgHeight) {
				uiMgr.setMessageLabel(null);
				return true;
			}
			if (is_border(x,y))
				return true;
			Superpixel s=get_superpixel(x,y);
			if (s!=hilightedSuperpixel) {
				boolean hasSameLabel = (s!=null && hilightedSuperpixel!=null && s.label()==hilightedSuperpixel.label());
				changeHilightedRegion(s);		
				if (keyIsDown && (scopeLabel==-2 || ((s==null && scopeLabel==-1) || (s!=null && s.label()==scopeLabel)))) {
					region_set_label(s);
					updateTexture(false);
				} else
					updateTexture(!hasSameLabel && !keyIsDown);
			}
			return true;
		} catch (java.lang.OutOfMemoryError e) {
			JOptionPane.showMessageDialog(null,"Out of memory. Switching to 3D pose","Memory Error",JOptionPane.ERROR_MESSAGE);
			uiMgr.setPose3DView(true);
			return false;
		}
	}

	private void changeSegmentationLevel(int step) {
		if (step>0) {
			if (segmentationLevel==255) return;
			segmentationLevel = Math.min(255, segmentationLevel+step);
		} else {
			if (segmentationLevel==0) return;
			segmentationLevel = Math.max(0, segmentationLevel+step);			
		}
		changeHilightedRegion(null);
		for (Superpixel s : superpixels)
			for (Border b : s.borders())
				b.update_boundary();
		updateTexture(true);
	}
	
	private void updateHilightedRegionOutline(int label) {
		for (Border b : hilightedRegionOutline)
			b.erase(leftDisplayImg);
		
		hilightedRegionOutline.clear();
		if (label!=-1) {
			// Collect all borders between superpixels with label s.label() and another label
			for (Superpixel p : superpixels)
				if (p.label()==label)
					for (Border b : p.borders())
						if (b.other(p).label()!=label)
							hilightedRegionOutline.add(b);
			
			for (Border b : hilightedRegionOutline)
				b.draw(leftDisplayImg, 0xFF0000);			
		}
	}
	
	private void changeHilightedRegion(Superpixel s) {
		if (s==hilightedSuperpixel)
			return;
		hilightedSuperpixel = s;

		updateHilightedRegionOutline(s==null ? -1 : s.label());

		for (Border b : hilightedOutline)
			b.set_hilighted(false);
		hilightedOutline.clear();
						
		if (s!=null) {
			s.get_region_outline(hilightedOutline);
			for (Border b : hilightedOutline)
				b.set_hilighted(true);
			
//			updateTexture(!hasSameLabel);
			if (s!=null && s.label()!=-1)
				uiMgr.setMessageLabel(regionNames[s.label()]+" ("+shortcuts[s.label()]+")");
			else
				uiMgr.setMessageLabel(null);
		} else
			uiMgr.setMessageLabel(null);
	}
	
	public boolean isRegionKeyShortcut(KeyEvent ev) {
		switch (ev.getKeyCode()) {
			case KeyEvent.VK_UP:
			case KeyEvent.VK_DOWN:
			case KeyEvent.VK_BACK_SPACE:
			case KeyEvent.VK_DELETE:
				return true;	
			case KeyEvent.VK_Z:
				return (ev.getModifiers() & KeyEvent.CTRL_MASK)>0;
				
		}
		for (int i=0; i<shortcuts.length; ++i)
			if (ev.getKeyChar()==shortcuts[i])
				return true;
		return false;
	}
	
	public void reset() {
		updateHilightedRegionOutline(-1);
		updateTexture(true);		
	}

	@Override
	protected boolean processKeyReleased(KeyEvent ev) {
		keyIsDown=false;
		scopeLabel=-2;
		updateTexture(true);
		return true;
	}

	@Override
	protected boolean processKeyPressed(KeyEvent ev) {
		if (suppressKeyEvents && uiMgr.keypointMgr.getKeypointFromShortcut(ev.getKeyChar())!=null)
			return false;
		if (keyIsDown)
			return true;
				
		switch (ev.getKeyCode()) {
			case KeyEvent.VK_LEFT:
			case KeyEvent.VK_RIGHT:
			case KeyEvent.VK_ENTER:
				uiMgr.processKeyPressed(ev);
				return true;
			case KeyEvent.VK_UP:
				changeSegmentationLevel(((ev.getModifiers() & KeyEvent.SHIFT_MASK)>0) ? 10 : 1);
				return true;
			case KeyEvent.VK_DOWN:
				changeSegmentationLevel(((ev.getModifiers() & KeyEvent.SHIFT_MASK)>0) ? -10 : -1);
				return true;
			case KeyEvent.VK_Z:
				if ((ev.getModifiers() & KeyEvent.CTRL_MASK)>0 && prevLabels!=null) {
					// CTRL-Z
					LinkedList<SuperpixelValuePair> newPrevLabels = new LinkedList<SuperpixelValuePair>();
					for (SuperpixelValuePair svp : prevLabels) {
						newPrevLabels.add(new SuperpixelValuePair(svp.s, svp.s.label()));
						svp.s.superpixel_set_label(svp.value);
					}
					updateHilightedRegionOutline(-1);	// Remove the hilighted region on the left window
					updateTexture(true);
					prevLabels = newPrevLabels;
				}
		}
		if (ev.getKeyChar() == ' ') {
			uiMgr.processKeyPressed(ev);
			return true;
		}
		
		currentLabel=-1;
		boolean delKeyDown = (ev.getKeyCode()==KeyEvent.VK_DELETE || ev.getKeyCode()==KeyEvent.VK_BACK_SPACE);
		if (!delKeyDown) {
			for (int i=0; i<shortcuts.length && currentLabel==-1; ++i)
				if (ev.getKeyChar()==shortcuts[i])
					currentLabel = i;
			if (currentLabel==-1)
				return true;	 // unrecognized key
		}

		keyIsDown = true;
		prevLabels = new LinkedList<SuperpixelValuePair>();

		Point2d coords = curImageCoords();
		Point2d pixCoords = unitToPixelCoords(coords.x,coords.y);

		int x=(int)Math.round(pixCoords.x);
		int y=(int)Math.round(pixCoords.y);

		if (x<0 && y>=0 && y<imgHeight) {
			uiMgr.processKeyPressed(ev);	// clicking to the left - give it to the keypoint UI
			return true;
		}

		Superpixel s = get_superpixel(x,y);
		
		if (s!=null) {
			scopeLabel = s.label();
			if (scopeLabel==currentLabel)
				scopeLabel=-2;
		}
//		System.out.println("Current label "+(currentLabel==-1 ? "DEL" : regionNames[currentLabel])+" Scope label:"+(scopeLabel<0 ? scopeLabel : regionNames[scopeLabel]));

		region_set_label(s);
		updateHilightedRegionOutline(-1);	// Remove the hilighted region on the left window
		updateTexture(true);
		return true;
	}
	
	private boolean region_set_label(Superpixel s) {
		if (s!=null) {			
			s.region_set_label(currentLabel);
			if (prevLabels==null)
				return false; // already set
			updateHilightedRegionOutline(currentLabel);
			if (s.label()!=-1)
				uiMgr.setMessageLabel(regionNames[s.label()]+" ("+shortcuts[s.label()]+")");
			else
				uiMgr.setMessageLabel(null);
		}
		
		uiMgr.setDirty(true);
		return true;
	}
	
	/*
	private void updateSubtexture(Rectangle bounds) {
		ImageComponent2D img = (ImageComponent2D)imageNode.getAppearance().getTexture().getImage(0);
		TextureLoader loader = new TextureLoader(displayImg.getSubimage(bounds.x, bounds.y, bounds.width, bounds.height));
		BufferedImage sc = ((ImageComponent2D)loader.getTexture().getImage(0)).getImage();
		double w = imgWidth/(double)img.getWidth();
		double h = imgHeight/(double)img.getHeight();
		img.setSubImage(sc, sc.getWidth(), sc.getHeight(), 0,0, (int)(bounds.x/w), (int)(bounds.y/h));		
	}
	*/
	private void updateTexture(boolean alsoUpdateLeftImage) {
//		ImageComponent2D img = (ImageComponent2D)imageNode.getAppearance().getTexture().getImage(0);
//		TextureLoader loader = new TextureLoader(displayImg.getSubimage(0, 0, imgWidth, imgHeight/2),TextureLoader.BY_REFERENCE);
//		img.setSubImage(((ImageComponent2D)loader.getTexture().getImage(0)).getImage(), imgWidth, imgHeight/2, 0,0, 0,0);

//		img.setSubImage(displayImg.getSubimage(0, 0, imgWidth, imgHeight), imgWidth, imgHeight, 0,0, 0,0);
//		img.setSubImage(displayImg.getSubimage(0, 0, imgWidth, imgHeight), imgWidth, imgHeight, 0,0, img.getWidth()-imgWidth,img.getHeight()-imgHeight);
//		imageNode.getAppearance().getTexture().setImage(0, img);

		imageNode.getAppearance().setTexture(null);	// to save memory
		TextureLoader loader = new TextureLoader(displayImg,"RGB");//,TextureLoader.BY_REFERENCE);
		loader.getTexture().setCapability(Texture.ALLOW_IMAGE_WRITE);
		loader.getTexture().getImage(0).setCapability(ImageComponent2D.ALLOW_IMAGE_WRITE);
		loader.getTexture().getImage(0).setCapability(ImageComponent2D.ALLOW_IMAGE_READ);
		imageNode.getAppearance().setTexture(loader.getTexture());		
		
		if (alsoUpdateLeftImage) {
			leftImageNode.getAppearance().setTexture(null);	// to save memory
			loader = new TextureLoader(leftDisplayImg,"RGB");//,TextureLoader.BY_REFERENCE);
			loader.getTexture().setCapability(Texture.ALLOW_IMAGE_WRITE);
			loader.getTexture().getImage(0).setCapability(ImageComponent2D.ALLOW_IMAGE_WRITE);
			loader.getTexture().getImage(0).setCapability(ImageComponent2D.ALLOW_IMAGE_READ);
			leftImageNode.getAppearance().setTexture(loader.getTexture());				
		}
	}
	public void enable(boolean enabled) {
		isEnabled= enabled;
		if (enabled) {
			canvas.requestFocus();
			if (superpixels==null)
				init();
			updateRegionLabels();
		} else {
			prevLabels = null;
		}
		keyIsDown=false;
		scopeLabel=-2;
		suppressKeyEvents = true;
	}
	public boolean isEnabled() {
		return isEnabled;
	}
	public void processStimulusComplete() {
		suppressKeyEvents = false;
	}
}
