/**
 * 
 */
package core;

import java.awt.Button;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Panel;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.media.j3d.Appearance;
import javax.media.j3d.BoundingBox;
import javax.media.j3d.Bounds;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Canvas3D;
import javax.media.j3d.ColoringAttributes;
import javax.media.j3d.GeometryArray;
import javax.media.j3d.LineArray;
import javax.media.j3d.RenderingAttributes;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Switch;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.swing.ButtonGroup;
import javax.swing.JFileChooser;
import javax.swing.JFormattedTextField;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JRadioButton;
import javax.swing.filechooser.FileFilter;
import javax.vecmath.Point2d;
import javax.vecmath.Point3d;

import com.sun.j3d.utils.picking.PickResult;
import com.sun.j3d.utils.universe.ViewingPlatform;

class UIManager extends PanZoomMgr implements ActionListener, PropertyChangeListener {
	private Keypoint hilightedKeypoint = null;
	KeypointMgr keypointMgr;
	private Shape3D activeSegment;
	private IOMgr ioMgr;
	private Human3DPose poseMgr;
	private SegmentationMgr segMgr;
	private ImageMgr imgMgr;
	private ConfigMgr configMgr;
	private ViewingPlatform viewPlatform;
	private Switch taskSwitch;
	private Button saveButton = new Button("Save");
	private Button newButton = new Button("New Entry");
	private Button delButton = new Button("Delete Entry");
    private JRadioButton[] subcategoryButtons;
    private JRadioButton unknownBody = new JRadioButton("None");    
    
    private JRadioButton pose3DViewButton = new JRadioButton("Pose3D");
    private JRadioButton segViewButton = new JRadioButton("Segmentation");
    
    private JFormattedTextField entryID;
	private JLabel statusLabel = new JLabel();
	private JLabel msgLabel = new JLabel();
	
	private boolean curEntryIsNew = false;
	private boolean is_dirty = false;

	public void setPoseMgr(Human3DPose poseMgr) { this.poseMgr = poseMgr; }
	
	public UIManager(Canvas3D canvas, BranchGroup root, Bounds bounds, 
			TransformGroup transitionXform1, TransformGroup transitionXform2, TransformGroup img_xform, TransformGroup pbImageXform,
			KeypointMgr keypoints, ViewingPlatform viewPlatform, ImageMgr imgMgr, IOMgr ioMgr, ConfigMgr configMgr, Switch taskSwitch) {
		super(canvas, root, bounds, transitionXform1, transitionXform2, img_xform, pbImageXform);
		this.keypointMgr = keypoints;
		this.imgMgr = imgMgr;
		this.viewPlatform = viewPlatform;
		this.ioMgr = ioMgr;
		this.configMgr = configMgr;
		this.taskSwitch = taskSwitch;
		this.taskSwitch.setCapability(Switch.ALLOW_SWITCH_READ);
		this.taskSwitch.setCapability(Switch.ALLOW_SWITCH_WRITE);
		
		// Create the active segment
		LineArray xGeom = new LineArray(4, GeometryArray.COORDINATES);
		xGeom.setCapability(LineArray.ALLOW_COORDINATE_WRITE);

		activeSegment =  new Shape3D(xGeom);
		Appearance ap = new Appearance();
		ap.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_WRITE);
		ap.setCapability(Appearance.ALLOW_RENDERING_ATTRIBUTES_WRITE);
		ap.setCapability(Appearance.ALLOW_RENDERING_ATTRIBUTES_READ);
		ColoringAttributes ca = new ColoringAttributes();
		ca.setColor(0,0,1);
		ap.setColoringAttributes(ca);
		RenderingAttributes ra = new RenderingAttributes();
		ra.setVisible(false);
		ra.setCapability(RenderingAttributes.ALLOW_VISIBLE_WRITE);
		ap.setRenderingAttributes(ra);
		activeSegment.setAppearance(ap);
		activeSegment.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
		activeSegment.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
		activeSegment.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
		keypoints.getRoot().addChild(activeSegment);		
	}	
	
	public void setSegMgr(SegmentationMgr segMgr) {
		this.segMgr = segMgr;
	}
	
	private boolean abandonChanges() {
		Object[] options = {"Yes","No", "Cancel"};
		int response = JOptionPane.showOptionDialog(null, "Do you want to save changes?",
			    "Switching to another annotation",
			    JOptionPane.YES_NO_CANCEL_OPTION,
			    JOptionPane.QUESTION_MESSAGE,
			    null,
			    options,
			    options[2]);

        if (response==0)
        	saveButtonPressed();
        return (response!=2);
	}
	
	public boolean processKeyPressed(KeyEvent ev) {
		if (!isInPose3DView() && segMgr.isRegionKeyShortcut(ev))
			return segMgr.processKeyPressed(ev);
		
		if (ev.getKeyChar()!='0' && ev.getKeyChar()!='9')
			poseMgr.endIncrementalCorrections();
		switch (ev.getKeyCode()) {
		case KeyEvent.VK_LEFT:
			if (ioMgr.curEntry()>0) {
				if (is_dirty && !abandonChanges())
					return true;
				
				boolean success=ioMgr.prevEntry();
				assert(success);
				syncFromIOMgr();
				poseMgr.resetView();
				centerData(true);
				setCurEntryIsNew(false);
				if (!isInPose3DView())
					segMgr.reset();
			}
			return false;
		case KeyEvent.VK_RIGHT:
			if (ioMgr.curEntry()+1<ioMgr.numEntries()) {
				if (is_dirty && !abandonChanges())
					return true;
				boolean success=ioMgr.nextEntry();
				assert(success);
				syncFromIOMgr();
				poseMgr.resetView();
				centerData(true);
				setCurEntryIsNew(false);
				if (!isInPose3DView())
					segMgr.reset();
			}
			return false;
		case KeyEvent.VK_DELETE:
		case KeyEvent.VK_BACK_SPACE:
			if (hilightedKeypoint!=null) {
				hilightedKeypoint.changeStatus(Keypoint.StatusEnum.Inactive);
				hilightedKeypoint=null;
				updateActiveSegment();
				poseMgr.display();
			}
			
			return true;
		case KeyEvent.VK_ENTER:
			if (saveButton.isEnabled())
				saveButtonPressed();			
			return false;
		}

		switch (ev.getKeyChar()) {
			case 'b':
				if (hilightedKeypoint!=null && hilightedKeypoint.isActive()) {
					hilightedKeypoint.toggleZOrder();
					updateActiveSegment();
					poseMgr.display();
				}
				break;		
			case 'n':
				if (hilightedKeypoint!=null && hilightedKeypoint.isActive()) {
					hilightedKeypoint.setHidden(!hilightedKeypoint.getHidden());
				}
				break;		
			case ' ':
				poseMgr.resetView();
				centerData(true);
				if (!isInPose3DView())
					segMgr.reset();
				break;
			case '8':
				poseMgr.fixTorso();
				poseMgr.display();
				break;
			case '9':
				poseMgr.doIncrementalCorrections(true,  false);
				poseMgr.display();
				break;
			case '(':
				poseMgr.doIncrementalCorrections(true,  true);
				poseMgr.display();
				break;
			case '0':
				poseMgr.doIncrementalCorrections(false, false);
				poseMgr.display();
				break;
			case ')':
				poseMgr.doIncrementalCorrections(false, true);
				poseMgr.display();
				break;
			default:
				if (isInPose3DView()) {
					Point2d img_pt = curImageCoords();
					keypointMgr.keyPressed(ev.getKeyChar(), img_pt.x,img_pt.y);
					poseMgr.display();
				}
		}
		return true;
	}
	public boolean processKeyReleased(KeyEvent ev) { 
		if (!isInPose3DView() && segMgr.isRegionKeyShortcut(ev))
			return segMgr.processKeyReleased(ev);
		return true;
	}
	 
	
	public boolean isReadOnly() { return ioMgr.isReadOnly(); }

	// The data in ioMgr has changed. Update the keypoints and image accordingly
	public void syncFromIOMgr() {
		boolean newImage = imgMgr.setImage(ioMgr.getCurImageFullName(), ioMgr.isReadOnly());
		if (newImage)
			segMgr.setImage(ioMgr.getCurPbImageFullName(), imgMgr.getBufferedImage(), imgMgr.getImage());
		keypointMgr.read(ioMgr);
		if (segMgr.hasSegments()) {
			segMgr.setRegionLabels(ioMgr.getRegionLabels());
			pose3DViewButton.setEnabled(true);
			segViewButton.setEnabled(true);
		} else {
			pose3DViewButton.setEnabled(false);
			segViewButton.setEnabled(false);
			setPose3DView(true);
			pose3DViewButton.setSelected(true);
		}
		viewPlatform.getViewPlatformTransform().getTransform(poseMgr.normalizedViewXform);
		updateLabelsScale();
		if (ioMgr.getBodyType()==-1) {
			unknownBody.setSelected(true);
		} else
			subcategoryButtons[ioMgr.getBodyType()].setSelected(true);
		setBodyType(ioMgr.getBodyType());
		poseMgr.resetView();
		centerData(!newImage);
		setDirty(false);
		statusLabel.setText((ioMgr.isReadOnly() ? "DEMO " : "")+ioMgr.getCurImageName(false)+" "+ioMgr.getCurEntryInImageIndex());
		entryID.setValue(ioMgr.curEntry()+1);
	}
		
	protected boolean processMouseMoved(MouseEvent mev) {
		super.processMouseMoved(mev);
		// Hilighting the active keypoint
		pickCanvas.setShapeLocation(mev.getX(), mev.getY());
		
		PickResult[] pickInfo = pickCanvas.pickAll();
		if (pickInfo!=null)							
			for (int p=0; p<pickInfo.length; ++p) {							
				Shape3D shape = (Shape3D) pickInfo[p].getObject();
				if (shape==null) continue;
			
				if (!(shape.getUserData() instanceof Keypoint))
					continue;

				setActivePoint((Keypoint)shape.getUserData());
				return true;
			}
		
		setActivePoint(null);
		return true;
	}
	
	private void setActivePoint(Keypoint k) {
		if (hilightedKeypoint == k)
			return;
				
		if (k!=null && k.isActive())
			k.changeStatus(Keypoint.StatusEnum.Hilighted);
		if (hilightedKeypoint!=null)
			hilightedKeypoint.changeStatus(Keypoint.StatusEnum.Active);
		if (k==null || k.isActive())
			hilightedKeypoint = k;	
		
		updateActiveSegment();
	}
	
	private void updateActiveSegment() {
		if (hilightedKeypoint!=null && hilightedKeypoint.isActive()) {
			Keypoint parent = keypointMgr.getKeypointIfActive(hilightedKeypoint.getParent());
			if (parent!=null) {
				LineArray xGeom = (LineArray)activeSegment.getGeometry();
				double Width = 0.04;
				
				double nearW = 0, farW=0;
				switch (hilightedKeypoint.getZOrder()) {
					case Close: nearW=Width*hilightedKeypoint.getLabelScale(); break;
					case Far: farW=Width*parent.getLabelScale(); break;
				}

				xGeom.setCoordinate(0, new Point3d(hilightedKeypoint.x(),hilightedKeypoint.y()+nearW,KeypointMgr.keypoints_z));
				xGeom.setCoordinate(1, new Point3d(parent.x(),parent.y()+farW,KeypointMgr.keypoints_z));
				xGeom.setCoordinate(2, new Point3d(hilightedKeypoint.x(),hilightedKeypoint.y()-nearW,KeypointMgr.keypoints_z));
				xGeom.setCoordinate(3, new Point3d(parent.x(),parent.y()-farW,KeypointMgr.keypoints_z));

				activeSegment.getAppearance().getRenderingAttributes().setVisible(true);
				return;
			}
		}
		activeSegment.getAppearance().getRenderingAttributes().setVisible(false);		
	}
	
	protected boolean processMouseDragged(MouseEvent mev) {		
		// using getButton doesn't work on Windows...
		boolean button3Down = (mev.getModifiers() & MouseEvent.BUTTON3_MASK)!=0;
							
		if (hilightedKeypoint!=null && !button3Down) {
			Point2d pt = screenToImagePlane(mev.getX(),mev.getY());
			hilightedKeypoint.moveTo(pt.x,pt.y);		
			hilightedKeypoint.changeStatus(Keypoint.StatusEnum.Hilighted);
			updateActiveSegment();
			poseMgr.endIncrementalCorrections();
			poseMgr.display();
			return true;
		} else
			return super.processMouseDragged(mev); // Move or zoom the image
	}
	
	private void updateLabelsScale() {
		Transform3D currXform = new Transform3D();
		img_xform.getTransform(currXform);
		keypointMgr.setLabelScale(0.5/currXform.getScale());
	}
		
	// scroll and zoom so the active keypoints are in full view	
	public void centerData(boolean doTransition) {
		BoundingBox b=keypointMgr.getBounds();
		segMgr.resetView(b, doTransition, false);
		resetView(b, doTransition, true);
		updateLabelsScale();
	}

	public boolean isDirty() { return is_dirty; }
	public void setDirty(boolean isDirty) {
		is_dirty = isDirty;
		saveButton.setEnabled(isDirty);
		saveButton.setForeground(isDirty ? new Color(255,0,0) : new Color(0,0,0));
	}
	
	public void setMessageLabel(String msg) {
		msgLabel.setText(msg);
	}
	
	public void setCurEntryIsNew(boolean isNew) {
		if (curEntryIsNew==isNew)
			return;
		curEntryIsNew=isNew;
		
		if (isNew)
			saveButton.setLabel("Save New Entry");
		else
			saveButton.setLabel("Save");
		saveButton.getParent().invalidate();
		saveButton.getParent().validate();
	}
	
	public void actionPerformed(ActionEvent ev) {
		if (ev.getSource()==saveButton) {
			saveButtonPressed();
		} else if (ev.getSource()==newButton) {
			ioMgr.clearKeypoints();
			syncFromIOMgr();
			centerData(true);
			setCurEntryIsNew(true);
		} else if (ev.getSource()==delButton) {
			if (ioMgr.deleteCurrentKeypointSet()) {
				delButton.setEnabled(ioMgr.numEntries()>1);
				syncFromIOMgr();
			}
		} else {
			for (int i=0; i<subcategoryButtons.length; ++i)				
				if (ev.getSource()==subcategoryButtons[i]) {
					setBodyType(i);
					return;
				}
		    if (ev.getSource()==pose3DViewButton)
			   setPose3DView(true);
		    else if (ev.getSource()==segViewButton)
			   setPose3DView(false);
		}
	}
	
	private boolean isInPose3DView() {
		return taskSwitch.getWhichChild()==0;
	}
	public void setPose3DView(boolean pose3DView) { // switches between 3D pose and segmentation
		poseMgr.enable(pose3DView);
		segMgr.enable(!pose3DView);
		taskSwitch.setWhichChild(pose3DView ? 0 : 1);
		pose3DViewButton.setSelected(pose3DView);
		segViewButton.setSelected(!pose3DView);
		
		if (!pose3DView) {
			poseMgr.resetView();
			centerData(true);
		} else
			canvas.requestFocus();
		keypointMgr.setVisible(pose3DView);
	}

	public void propertyChange(PropertyChangeEvent ev) {
		if (ev.getSource()==entryID) {
			Integer val = (Integer) entryID.getValue();
			if (val<1 || val>ioMgr.numEntries()) {
				if (val<1)
					val = 1;
				else
					val=ioMgr.numEntries();
				entryID.setValue(val);
			}
			canvas.requestFocus();
			if (is_dirty && !abandonChanges())
				return;

			if (ioMgr.gotoEntry(val-1)) {
				syncFromIOMgr();
				poseMgr.resetView();
				centerData(true);
				setCurEntryIsNew(false);
			}
		}
	}
	
	private void setBodyType(int bodyType) {
		keypointMgr.setBodyType(bodyType);
		poseMgr.setBodyType(bodyType);
		setDirty(true);
		poseMgr.display();
	}
	
	private void saveButtonPressed() {
		System.out.println("Save");
		keypointMgr.write(ioMgr);
		if (segMgr.hasSegments()) {
			boolean segIsEnabled=segMgr.isEnabled();
			segMgr.enable(true);	// this will init the region labels if not done already
			ioMgr.setRegionLabels(segMgr.getRegionLabels());
			segMgr.enable(segIsEnabled);
		}
		boolean success = (curEntryIsNew ? ioMgr.writeKeypointsInANewSet() : ioMgr.writeKeypoints());
		if (success) {
			setDirty(false);
			setCurEntryIsNew(false);
		}
	}
	
	public void addButtonPanel(Panel buttonPanel) {
		buttonPanel.setFocusable(false);
		
		buttonPanel.add(statusLabel);
		statusLabel.setPreferredSize(new Dimension(300,16));
		
		saveButton.setFocusable(false);
		buttonPanel.add(saveButton);
		saveButton.addActionListener(this);
		setCurEntryIsNew(false);

		newButton.setFocusable(false);
		buttonPanel.add(newButton);
		newButton.addActionListener(this);
		
		delButton.setFocusable(false);
		buttonPanel.add(delButton);
		delButton.setEnabled(ioMgr.numEntries()>1);
		delButton.addActionListener(this);
		
		ButtonGroup  bodyTypeButtonGroup = new ButtonGroup();
		subcategoryButtons = new JRadioButton[configMgr.getSubcategoryNames().length];
		for (int i=0; i<configMgr.getSubcategoryNames().length; ++i) {
			subcategoryButtons[i]= new JRadioButton(configMgr.getSubcategoryNames()[i]);
			subcategoryButtons[i].addActionListener(this);
			subcategoryButtons[i].setFocusable(false);
		    bodyTypeButtonGroup.add(subcategoryButtons[i]);
			buttonPanel.add(subcategoryButtons[i]);			
		}
					
		bodyTypeButtonGroup.add(unknownBody);		
		
		buttonPanel.add(new JLabel(" Current Entry:"));
	    entryID = new JFormattedTextField(new Integer(0));
	    entryID.setColumns(3);
	    entryID.setToolTipText("Double-click to change");
	    buttonPanel.add(entryID);	
	    entryID.addPropertyChangeListener("value",this);

	    ButtonGroup  viewButtonGroup = new ButtonGroup();

		pose3DViewButton.addActionListener(this);
		pose3DViewButton.setFocusable(false);
		viewButtonGroup.add(pose3DViewButton);
		buttonPanel.add(pose3DViewButton);
		pose3DViewButton.setSelected(true);
				
		segViewButton.addActionListener(this);
		segViewButton.setFocusable(false);
		viewButtonGroup.add(segViewButton);
		buttonPanel.add(segViewButton);

		buttonPanel.add(msgLabel);
		msgLabel.setPreferredSize(new Dimension(120,16));
	}
}